#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2006-2016 University of Dundee. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# @author Donald MacDonald &nbsp;&nbsp;&nbsp;&nbsp;
# <a href="mailto:donald@lifesci.dundee.ac.uk">donald@lifesci.dundee.ac.uk</a>

from __future__ import division
from __future__ import print_function

from builtins import range
from past.utils import old_div
from builtins import object
import math
import numpy

from omero.model import ImageI
from omero.model import RoiI
from omero.model import EllipseI
from omero.model import RectangleI
from omero.model import PolygonI
from omero.model import MaskI
from omero.model import NamespaceI
from omero.rtypes import rdouble, rint, rlong, rstring

import warnings
# Popo helpers #


def toCSV(list):
    """
    Convert a list to a Comma Separated Value string.
    @param list The list to convert.
    @return See above.
    """
    warnings.warn(
        "This module is deprecated as of OMERO 5.3.0", DeprecationWarning)
    lenList = len(list)
    cnt = 0
    str = ""
    for item in list:
        str = str + item
        if(cnt < lenList - 1):
            str = str + ","
        cnt = cnt + 1
    return str


def toList(csvString):
    """
    Convert a csv string to a list of strings
    @param csvString The CSV string to convert.
    @return See above.
    """
    warnings.warn(
        "This module is deprecated as of OMERO 5.3.0", DeprecationWarning)
    list = csvString.split(',')
    for index in range(len(list)):
        list[index] = list[index].strip()
    return list

##
# Create instance of data object this object wraps the basic OMERO types.
#


class DataObject(object):

    ##
    # Create instance.
    #

    def __init__(self):
        warnings.warn(
            "This module is deprecated as of OMERO 5.3.0", DeprecationWarning)
        self.value = None
        self.dirty = False

    ##
    # Sets the {@link IObject}.
    #
    # @param value The value to set.
    #
    def setValue(self, value):
        if(value is None):
            raise Exception("IObject delegate for DataObject cannot be null.")
        self.value = value

    ##
    # get the id of the Dataobject.
    # @return See above.
    #
    def getId(self):
        if(self.value.getId() is None):
            return -1
        return self.value.getId().getValue()

    ##
    # Set the id of the data object
    # @param id See above.
    def setId(self, id):
        self.setDirty(True)
        self.value.setId(rlong(id))

    ##
    # Get the current object.
    # @return See above.
    #
    def asIObject(self):
        return self.value

    ##
    # The object has been changed and is now dirty.
    # @param boolean See above.
    #
    def setDirty(self, boolean):
        self.dirty = boolean

    ##
    # Has the object has been changed.
    # @return See above.
    #
    def getDirty(self):
        return self.dirty

    ##
    # Is the object loaded
    # @return see above.
    def isLoaded(self):
        return self.value.isLoaded()

    ##
    # Get the user details for the object.
    # @return see above.
    def getDetails(self):
        return self.asIObject().getDetails()


class ImageData(DataObject):

    ##
    # Create Instance
    #

    def __init__(self, image=None):
        DataObject.__init__(self)
        if(image is None):
            self.setValue(ImageI())
        else:
            self.setValue(image)

    ##
    # Sets the name of the image.
    #
    # @param name
    # The name of the image. Mustn't be <code>null</code>.
    #
    def setName(self, name):
        image = self.asIObject()
        if(image is None):
            raise Exception("No Image specified.")
        image.setName(rstring(name))
        self.setDirty(True)

    ##
    # Returns the name of the image.
    #
    # @return See above.
    #
    def getName(self):
        image = self.asIObject()
        if(image is None):
            raise Exception("No Image specified.")
        name = image.getName()
        if(name is None):
            return ""
        return name.getValue()

    ##
    # Sets the description of the image.
    #
    # @param description
    #            The description of the image.
    #
    def setDescription(self, description):
        image = self.asIObject()
        if(image is None):
            raise Exception("No Image specified.")
        image.setDescription(rstring(description))
        self.setDirty(True)

    ##
    # Returns the description of the image.
    #
    # @return See above.
    #
    def getDescription(self):
        image = self.asIObject()
        if(image is None):
            raise Exception("No Image specified.")
        description = image.getDescription()
        if(description is None):
            return ""
        return description.getValue()

##
# This class stores the ROI Coordinate (Z,T).
#


class ROICoordinate(object):

    ##
    # Initialise the ROICoordinate.
    # @param z The z-section.
    # @param t The timepoint.

    def __init__(self, z=0, t=0):
        self.theZ = z
        self.theT = t
        self.ZBITSPLIT = 18

    ##
    # Overload the equals operator
    #
    def __eq__(self, obj):
        if(self.theZ == obj.theZ and self.theT == obj.theT):
            return True
        return False

    ##
    # Overload the equals operator
    #
    def __ne__(self, obj):
        if(self.theZ != obj.theZ or self.theT != obj.theT):
            return True
        return False

    ##
    # Overload the lessthan or equals operator
    #
    def __lt__(self, obj):
        if(self.theT >= obj.theT):
            return False
        if(self.theZ >= obj.theZ):
            return False
        return True

    ##
    # Overload the lessthan or equals operator
    #
    def __le__(self, obj):
        if(self.theT < obj.theT):
            return False
        if(self.theZ < obj.theZ):
            return False
        return True

    ##
    # Overload the greater than equals operator
    #
    def __gt__(self, obj):
        if(self.theT <= obj.theT):
            return False
        if(self.theZ <= obj.theZ):
            return False
        return True

    ##
    # Overload the greater than or equals operator
    #
    def __ge__(self, obj):
        if(self.theT < obj.theT):
            return False
        if(self.theZ < obj.theZ):
            return False
        return True

    ##
    # Overload the hash operator
    #
    def __hash__(self):
        return self.theZ << self.ZBITSPLIT + self.theT

    ##
    # Returns the timepoint.
    #
    # @return See above.
    #
    def getTimepoint(self):
        return self.theT

    ##
    # Returns the Z-Section.
    #
    # @return See above.
    #
    def getZSection(self):
        return self.theZ

    ##
    # Set the Z-Section of the Coordinate
    # @param z See above.
    #
    def setZSection(self, z):
        self.z = z

    ##
    # Set the Timepoint of the Coordinate
    # @param t See above.
    #
    def setTimepoint(self, t):
        self.t = t


###
# Shape wrapper.
#
def shapeWrapper(serverSideShape):
    """
    Wrap the serverSide shape as the appropriate OmeroPopos
    @param serverSideShape The shape object to wrap.
    @return See above.
    """
    print("ServerSideShape")
    print(serverSideShape.__class__.__name__)
    if serverSideShape.__class__.__name__ == 'EllipseI':
        return EllipseData(serverSideShape)
    if serverSideShape.__class__.__name__ == 'RectangleI':
        return RectData(serverSideShape)
    if serverSideShape.__class__.__name__ == 'MaskI':
        return MaskData(serverSideShape)
    if serverSideShape.__class__.__name__ == 'PolygonI':
        return PolygonData(serverSideShape)
    return None

##
# This class defines the python mapping of
# the ROIData object {See Pojos#ROIData}
#


class ROIData(DataObject):

    ##
    # Create a new instance of an ROIData object.
    #

    def __init__(self, roi=None):
        DataObject.__init__(self)
        if(roi is None):
            self.setValue(RoiI())
        else:
            self.setValue(roi)
        self.roiShapes = dict()
        if(roi is not None):
            self.initialise()
    ##
    # Initialise the shape map of the ROIData object.
    #

    def initialise(self):
        self.roiShapes = dict()
        roi = self.asIObject()
        shapes = roi.copyShapes()
        s = None
        for shape in shapes:
            s = shapeWrapper(shape)
            if(s is not None):
                coord = ROICoordinate(s.getZ(), s.getT())
                if(coord not in list(self.roiShapes.keys())):
                    self.roiShapes[coord] = list()
                data = self.roiShapes[coord]
                data.append(s)

    ##
    # Set the imageId for the ROI.
    # @param imageId See above.
    #
    def setImage(self, image):
        roi = self.asIObject()
        if(roi is None):
            raise Exception("No Roi specified.")
        roi.setImage(image)
        self.setDirty(True)

    ##
    # Get the image for the ROI.
    # @return See above.
    #
    def getImage(self):
        roi = self.asIObject()
        if(roi is None):
            raise Exception("No Roi specified.")
        return roi.getImage()

    ##
    # Add ShapeData object to ROIData.
    # @param shape See above.
    #
    def addShapeData(self, shape):
        roi = self.asIObject()
        if(roi is None):
            raise Exception("No Roi specified.")
        coord = shape.getROICoordinate()
        shapeList = None
        if(coord not in list(self.roiShapes.keys())):
            shapeList = list()
            self.roiShapes[coord] = shapeList
        else:
            shapeList = self.roiShapes[coord]
        shapeList.append(shape)
        roi.addShape(shape.asIObject())
        self.setDirty(True)

    ##
    # Remove ShapeData object from ROIData.
    # @param shape See above.
    #
    def removeShapeData(self, shape):
        roi = self.asIObject()
        if(roi is None):
            raise Exception("No Roi specified.")
        coord = shape.getROICoordinate()
        shapeList = self.roiShapes[coord]
        shapeList.remove(shape)
        roi.removeShape(shape.asIObject())
        self.setDirty(True)

    ##
    # Get the number of planes occupied by the ROI.
    # @return See above.
    #
    def getPlaneCount(self):
        return len(self.roiShapes)

    ##
    # Get the number of shapes in the ROI.
    # @return See above.
    #
    def getShapeCount(self):
        count = 0
        for coord in self.roiShapes:
            list = self.roiShapes[coord]
            count = count + len(list)
        return count

    ##
    # Returns the list of shapes on a given plane.
    #
    # @param z The z-section.
    # @param t The timepoint.
    # @return See above.
    #
    def getShapes(self, z, t):
        return self.roiShapes[ROICoordinate(z, t)]

    ##
    # Returns the iterator of the collection of the map.
    #
    # @return See above.
    #
    def getIterator(self):
        return iter(self.roiShapes.items())

    ##
    # Returns an iterator of the Shapes in the ROI in the range [start, end].
    #
    # @param start The starting plane where the Shapes should reside.
    # @param end The final plane where the Shapes should reside.
    # @return See above.
    #
    def getShapesInRange(self, start, end):
        coordList = list(self.roiShapes.keys())
        coordList.sort()
        keyList = []
        for coord in coordList:
            if(coord >= start and coord <= end):
                keyList.append(coord)
        return self.roiShapes.from_keys(keyList)


class ShapeData(DataObject):

    def __init__(self):
        DataObject.__init__(self)
        self.text = None
        self.coord = ROICoordinate()

    ##
    # Returns the z-section.
    #
    # @return See above.
    #
    def getZ(self):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        z = shape.getTheZ()
        if(z is None):
            return 0
        else:
            return z.getValue()

    ##
    # Set the z-section.
    # @param theZ See above.
    #
    def setZ(self, theZ):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        shape.setTheZ(rint(theZ))
        self.coord.setZSection(theZ)
        self.setDirty(True)

    ##
    # Returns the timepoint.
    #
    # @return See above.
    #
    def getT(self):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        t = shape.getTheT()
        if(t is None):
            return 0
        else:
            return t.getValue()

    ##
    # Set the timepoint.
    # @param See above.
    #
    def setT(self, theT):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        shape.setTheT(rint(theT))
        self.coord.setTimePoint(theT)
        self.setDirty(True)

    ##
    # Set the ROICoordinate for the ShapeData
    # @param roiCoordinate See above.
    #
    def setROICoordinate(self, coord):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        self.setZ(coord.getZSection())
        self.setT(coord.getTimePoint())
        self.coord.setZSection(coord.getZSection())
        self.coord.setTimePoint(coord.getTimePoint())
        self.setDirty(True)

    ##
    # Get the ROICoordinate for the ShapeData
    # @return See above.
    #
    def getROICoordinate(self):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        return self.coord

    ##
    # Get the text for the Object
    # @return See above.
    def getText(self):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        text = shape.getTextValue()
        if(text is None):
            return ""
        else:
            return text.getValue()

    ##
    # Set the text for the Obect.
    # @param See above.
    def setText(self, text):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        shape.setTextValue(rstring(text))
        self.setDirty(True)

    ##
    # Get the affinetransform from the object,
    # returned as a string matrix(m00 m01 m10 m11 m02 m12)
    # see Java affinetransform toMatrix.
    # @return see above.
    #
    def getTransform(self):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        transform = shape.getTransform()
        if(transform is not None):
            transformValue = transform.getValue()
            if(transformValue == "none"):
                return ""
            else:
                return transformValue
        return ""

    def setTransform(self, transform):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        shape.setTransform(rstring(transform))
        self.setDirty(True)

    ##
    # Transform the affine transform matrix from the string
    # 'matrix(m00 m01 m10 m 11 m02 m12)' to a more appropriate
    # numpy.array([m00 m01 m02], [m10 m11 m12]).
    #
    def transformToMatrix(self, str):
        if (str == ""):
            return numpy.matrix([[1, 0, 0], [0, 1, 0]])
        transformstr = str[str.find('(') + 1:len(str) - 1]
        values = transformstr.split(' ')
        b = numpy.matrix(numpy.array(values, dtype='double'))
        t = numpy.matrix(numpy.zeros((3, 3)))
        t[0, 0] = b[0, 0]
        t[0, 1] = b[0, 2]
        t[1, 0] = b[0, 1]
        t[1, 1] = b[0, 3]
        t[0, 2] = b[0, 4]
        t[1, 2] = b[0, 5]
        t[2, 2] = 1
        return t

    ##
    # does the shape contain the point
    # @return see above.
    #
    def contains(self, point):
        return False

    ##
    #
    #
    def containsPoints(self):
        return []

##
# Instance of the EllipseData Object
#


class EllipseData(ShapeData):

    ##
    # Create instance of EllipseData Object
    #

    def __init__(self, shape=None):
        ShapeData.__init__(self)
        if(shape is None):
            self.setValue(EllipseI())
            self.setX(0)
            self.setY(0)
            self.setRadiusX(0)
            self.setRadiusY(0)
        else:
            self.setValue(shape)

    ##
    # Set the centre x coord of the Ellipse
    # @param x See above.
    def setX(self, x):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        shape.setX(rdouble(x))

    ##
    # Get the centre x coord of the Ellipse
    # @return See Above.
    def getX(self):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        x = shape.getX()
        if(x is None):
            return 0
        return x.getValue()

    ##
    # Set the centre y coord of the Ellipse
    # @param y See above.
    def setY(self, y):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        shape.setY(rdouble(y))

    ##
    # Get the centre y coord of the Ellipse
    # @return See Above.
    def getY(self):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        y = shape.getY()
        if(y is None):
            return 0
        return y.getValue()

    ##
    # Set the radius on the x-axis of the Ellipse
    # @param radiusx See above.
    def setRadiusX(self, radiusx):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        shape.setRadiusX(rdouble(radiusx))

    ##
    # Get the radius of the x-axis of the Ellipse
    # @return See Above.
    def getRadiusX(self):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        radiusx = shape.getRadiusX()
        if(radiusx is None):
            return 0
        return radiusx.getValue()

    ##
    # Set the radius on the y-axis of the Ellipse
    # @param radiusy See above.
    def setRadiusY(self, radiusy):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        shape.setRadiusY(rdouble(radiusy))

    ##
    # Get the radius of the y-axis of the Ellipse
    # @return See Above.
    def getRadiusY(self):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        radiusy = shape.getRadiusY()
        if(radiusy is None):
            return 0
        return radiusy.getValue()

    ##
    # Transform the point by the affineTransform transform.
    # @param transform See above.
    # @param point See above.
    # @return See above.
    #
    def transformPoint(self, transform, point):
        p = numpy.matrix(point).transpose()
        return transform * p

    ##
    # Return a map of points(x,y) contained within the Shape
    # @return See above.
    #
    def containsPoints(self):
        x = self.getX()
        y = self.getY()
        radiusx = self.getRadiusX()
        radiusy = self.getRadiusY()
        transform = self.transformToMatrix(self.getTransform())
        point = numpy.matrix((x, y, 1)).transpose()
        centre = transform * point
        BL = numpy.matrix((x - radiusx, y + radiusy, 1)).transpose()
        BR = numpy.matrix((x + radiusx, y + radiusy, 1)).transpose()
        TL = numpy.matrix((x - radiusx, y - radiusy, 1)).transpose()
        TR = numpy.matrix((x + radiusx, y - radiusy, 1)).transpose()
        MajorAxisLeft = numpy.matrix((x - radiusx, y, 1)).transpose()
        MajorAxisRight = numpy.matrix((x + radiusx, y, 1)).transpose()
        lb = transform * BL
        rb = transform * BR
        lt = transform * TL
        rt = transform * TR
        majl = transform * MajorAxisLeft
        majr = transform * MajorAxisRight
        o = (majr[1] - majl[1])
        a = (majr[0] - majl[0])
        h = math.sqrt(o * o + a * a)
        majorAxisAngle = math.asin(old_div(o, h))
        boundingBoxMinX = min(lt[0], rt[0], lb[0], rb[0])
        boundingBoxMaxX = max(lt[0], rt[0], lb[0], rb[0])
        boundingBoxMinY = min(lt[1], rt[1], lb[1], rb[1])
        boundingBoxMaxY = max(lt[1], rt[1], lb[1], rb[1])
        boundingBox = (
            (boundingBoxMinX, boundingBoxMinY),
            (boundingBoxMaxX, boundingBoxMaxY))
        centredBoundingBox = (
            (boundingBox[0][0] - centre[0], boundingBox[0][1] - centre[1]),
            (boundingBox[1][0] - centre[0], boundingBox[1][1] - centre[1]))
        points = {}
        cx = float(centre[0])
        cy = float(centre[1])
        xrange = list(range(centredBoundingBox[0][0], centredBoundingBox[1][0]))
        yrange = list(range(centredBoundingBox[0][1], centredBoundingBox[1][1]))
        for dx in xrange:
            for dy in yrange:
                newX = dx * math.cos(majorAxisAngle) + dy * \
                    math.sin(majorAxisAngle)
                newY = -dx * math.sin(majorAxisAngle) + \
                    dy * math.cos(majorAxisAngle)
                val = old_div((newX * newX), (radiusx * radiusx)) + \
                    old_div((newY * newY), (radiusy * radiusy))
                if(val <= 1):
                    points[(int(dx + cx), int(dy + cy))] = 1
        return points

##
# Instance of Polygon object.
#


class PolygonData(ShapeData):

    ###
    # Create instance of PolygonData Object
    #

    def __init__(self, shape=None):
        ShapeData.__init__(self)
        self.NUMREGEX = "\\[.*\\]"
        # Regex for a data in block.
        if(shape is None):
            self.setValue(PolygonI())
            self.points = []
            self.points1 = []
            self.points2 = []
            self.mask = []
        else:
            self.setValue(shape)
            self.parseShapeStringToPoints()

    ##
    # Get the points from the points String
    # @return See above.
    #
    def getPoints(self):
        pts = self.fromPoints("points")
        return self.parsePointsToList(pts)

    ##
    # Get the points1 from the points String
    # @return See above.
    #
    def getPoints1(self):
        pts = self.fromPoints("points1")
        return self.parsePointsToList(pts)

    ##
    # Get the points2 from the points String
    # @return See above.
    #
    def getPoints2(self):
        pts = self.fromPoints("points2")
        return self.parsePointsToList(pts)

    ##
    # Get the mask type from the points String
    # @return See above.
    #
    def getMaskPoints(self):
        pts = self.fromPoints("mask")
        return self.parsePointsToList(pts)

    ##
    # Set the points from the original PolygonI type.
    # @param pts The points values.
    #
    def setPointsString(self, pts):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        shape.setPoints(pts)
        self.setDirty(True)
        self.parseShapeStringToPoints()

    ##
    # Set the points from a series of lists, and also set the points string.
    # @param points The points list.
    # @param points1 The points1 list.
    # @param points2 The points2 list.
    # @param mask The mask represents the curve type, lineTo, ArcTo..
    def setPointsFromList(self, points, points1, point2, mask):
        pts = self.toString(points)
        pts1 = self.toString(points)
        pts2 = self.toString(points)
        mask = self.toString(points)
        str = "points[" + pts + "] "
        str = str + "points1[" + pts1 + "] "
        str = str + "points2[" + pts2 + "] "
        str = str + "mask[" + mask + "]"
        self.setPointsString(str)

    ##
    # Get the points string from the IObject
    # @return See above.
    def getPointsString(self):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        pts = shape.getPoints()
        if(pts is None):
            return ""
        else:
            return pts.getValue()

    ##
    # Get the points of type from the point string.
    # @param type The points type to return, points, point1, point2, mask
    # @return See above.
    #
    def fromPoints(self, type):
        return self.getContents(self.getPointsString(), type + "[", "]")

    ##
    # Helper method to get the contents of the values
    # from the string of form type[values].
    # @param string The string to parse.
    # @param start the first part of the string to break apart on. "type["
    # @param end the last part of the string to break apart "]".
    # @return The contents of the string between start ReturnValues end
    #
    def getContents(self, string, start, end):
        lIndex = string.find(start)
        if(lIndex == -1):
            return ""
        strFragment = string[lIndex:]
        rIndex = strFragment.find(']')
        if(rIndex == -1):
            return ""
        return string[lIndex + len(start):rIndex]

    ##
    # Convert the pts string to a list
    # @return See above.
    #
    def parsePointsToList(self, pts):
        numberList = pts.split(',')
        return numberList

    ##
    # Convert the pointsList to a string, of CSV
    # @return See above.
    #
    def toString(self, pointsList):
        str = ""
        for index, pt in enumerate(pointsList):
            str = str + pt
            if(index < len(pointsList) - 1):
                str = str + ","
        return str

    ##
    # Convert the points string to the points lists,
    # points, point1, point2 and mask.
    #
    def parseShapeStringToPoints(self):
        self.points = self.fromPoints("points")
        self.points1 = self.fromPoints("points1")
        self.points2 = self.fromPoints("points2")
        self.mask = self.fromPoints("mask")

    ##
    # Returns the bounding rectangle of the polygon,
    # as a list of coords [(x1,y1), (x2,y2)]
    # @return See above.
    #
    def getBoundingRectangle(self):
        pts = self.toCoords(self.getPoints())
        minx = pts[0][0]
        maxx = minx
        miny = pts[0][1]
        maxy = miny

        for pt in pts:
            minx = min(pt[0], minx)
            miny = min(pt[1], miny)
            maxx = max(pt[0], maxx)
            maxy = max(pt[1], maxy)
        return [(minx, miny), (maxx, maxy)]

    ##
    # Convert the points list to a coord list.
    # @param ptsList The list of points.
    # @return See above.
    def toCoords(self, ptsList):
        coords = []
        for index in range(old_div(len(ptsList), 2)):
            coords.append(
                (int(ptsList[index * 2]), int(ptsList[index * 2 + 1])))
        return coords

    ##
    # Return a map of points(x,y) contained within the Shape
    # @return See above.
    #
    def containsPoints(self):
        points = {}
        boundingRectangle = self.getBoundingRectangle()
        xrange = list(range(boundingRectangle[0][0], boundingRectangle[1][0]))
        yrange = list(range(boundingRectangle[0][1], boundingRectangle[1][1]))
        for xx in xrange:
            for yy in yrange:
                if(self.inPolygon((xx, yy))):
                    points[(xx, yy)] = 1
        return points

    ##
    # Return true if the point p is inside the polygon,
    # defined by the vertexes in points.
    # @param p The point (x,y)
    # @return See above.
    #
    def inPolygon(self, p):
        angle = 0.0
        polypoints = self.getPoints()
        polygon = []
        for index in range(0, old_div(len(polypoints), 2)):
            polygon.append(
                (int(polypoints[index * 2]), int(polypoints[index * 2 + 1])))

        n = len(polygon)

        for i, (h, v) in enumerate(polygon):
            p1 = (h - p[0], v - p[1])
            h, v = polygon[(i + 1) % n]
            p2 = (h - p[0], v - p[1])
            angle += self.Angle2D(p1[0], p1[1], p2[0], p2[1])

        if abs(angle) < math.pi:
            return False
        return True

    ##
    # Return the angle(in radians) between the two vectors (x1,y1), (x2,y2)
    # @param x1 The x of the first vector
    # @param y1 The y of the first vector
    # @param x2 The x of the second vector
    # @param y2 The y of the second vector
    # @return see above.
    #
    def Angle2D(self,  x1, y1, x2, y2):
        theta1 = math.atan2(y1, x1)
        theta2 = math.atan2(y2, x2)
        dtheta = theta2 - theta1
        while dtheta > math.pi:
            dtheta -= 2.0 * math.pi
        while dtheta < -math.pi:
            dtheta += 2.0 * math.pi
        return dtheta

##
# Instance of the Mask Object
#


class MaskData(ShapeData):

    ##
    # Create instance of MaskData Object
    #

    def __init__(self, maskShape=None):
        ShapeData.__init__(self)
        if(maskShape is None):
            self.setValue(MaskI())
            self.setX(0)
            self.setY(0)
            self.setWidth(0)
            self.setHeight(0)
            self.setBytes(None)
        else:
            self.setValue(maskShape)

    ##
    # Set the x coord of the Mask
    # @param x See above.
    def setX(self, x):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        shape.setX(rdouble(x))

    ##
    # Get the x coord of the Mask
    # @return See Above.
    def getX(self):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        x = shape.getX()
        if(x is None):
            return 0
        return x.getValue()

    ##
    # Set the y coord of the Mask
    # @param y See above.
    def setY(self, y):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        shape.setY(rdouble(y))

    ##
    # Get the y coord of the Mask
    # @return See Above.
    def getY(self):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        y = shape.getY()
        if(y is None):
            return 0
        return y.getValue()
    ##
    # Set the width the Mask
    # @param width See above.

    def setWidth(self, width):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        shape.setWidth(rdouble(width))

    ##
    # Get the width of the Mask
    # @return See Above.
    def getWidth(self):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        width = shape.getWidth()
        if(width is None):
            return 0
        return width.getValue()
    ##
    # Set the height of the Mask
    # @param height See above.

    def setHeight(self, height):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        shape.setHeight(rdouble(height))

    ##
    # Get the height of the Mask
    # @return See Above.
    def getHeight(self):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        height = shape.getHeight()
        if(height is None):
            return 0
        return height.getValue()
    ##
    # Set the bitmask of the Mask
    # @param See Above.

    def setMask(self, mask):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        shape.setBytes(mask)

    ##
    # Get the bitmask of the Mask
    # @return See Above.
    def getMask(self):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        mask = shape.getBytes()
        if(mask is None):  # ??
            return 0
        return mask.getValue()


##
# Instance of the RectangleData object
#
class RectData(ShapeData):

    ##
    # Create instance of MaskData Object
    #

    def __init__(self, rectShape=None):
        ShapeData.__init__(self)
        if (rectShape is None):
            self.setValue(RectangleI())
            self.setX(0)
            self.setY(0)
            self.setWidth(0)
            self.setHeight(0)
        else:
            self.setValue(rectShape)

    ##
    # Set the x coord of the Rectangle
    # @param x See above.
    def setX(self, x):
        shape = self.asIObject()
        if (shape is None):
            raise Exception("No Shape specified.")
        shape.setX(rdouble(x))

    ##
    # Get the x coord of the Rectangle
    # @return See Above.
    def getX(self):
        shape = self.asIObject()
        if (shape is None):
            raise Exception("No Shape specified.")
        x = shape.getX()
        if (x is None):
            return 0
        return x.getValue()

    ##
    # Set the y coord of the Rectangle
    # @param y See above.
    def setY(self, y):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        shape.setY(rdouble(y))

    ##
    # Get the y coord of the Rectangle
    # @return See Above.
    def getY(self):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        y = shape.getY()
        if(y is None):
            return 0
        return y.getValue()
    ##
    # Set the width the Rectangle
    # @param width See above.

    def setWidth(self, width):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        shape.setWidth(rdouble(width))

    ##
    # Get the width of the Rectangle
    # @return See Above.
    def getWidth(self):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        width = shape.getWidth()
        if(width is None):
            return 0
        return width.getValue()
    ##
    # Set the height of the Rectangle
    # @param height See above.

    def setHeight(self, height):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        shape.setHeight(rdouble(height))

    ##
    # Get the height of the Rectangle
    # @return See Above.
    def getHeight(self):
        shape = self.asIObject()
        if(shape is None):
            raise Exception("No Shape specified.")
        height = shape.getHeight()
        if(height is None):
            return 0
        return height.getValue()

    ##
    # Transform the point by the affineTransform transform.
    # @param transform See above.
    # @param point See above.
    # @return See above.
    #
    def transformPoint(self, transform, point):
        p = numpy.matrix(point).transpose()
        return transform * p

    ##
    # Return a map of points(x,y) contained within the Shape
    # @return See above.
    #
    def containsPoints(self):
        transform = self.transformToMatrix(self.getTransform())
        x = self.getX()
        y = self.getY()
        width = self.getWidth()
        height = self.getHeight()
        point = numpy.matrix((x, y, 1)).transpose()
        centre = transform * point
        BL = numpy.matrix((x, y + height, 1)).transpose()
        BR = numpy.matrix((x + width, y + height, 1)).transpose()
        TL = numpy.matrix((x, y, 1)).transpose()
        TR = numpy.matrix((x + width, y, 1)).transpose()
        lb = transform * BL
        rb = transform * BR
        lt = transform * TL
        rt = transform * TR
        majl = lb
        majr = rb
        o = (majr[1] - majl[1])
        a = (majr[0] - majl[0])
        h = math.sqrt(o * o + a * a)
        angle = math.asin(old_div(o, h))
        boundingBoxMinX = min(lt[0], rt[0], lb[0], rb[0])
        boundingBoxMaxX = max(lt[0], rt[0], lb[0], rb[0])
        boundingBoxMinY = min(lt[1], rt[1], lb[1], rb[1])
        boundingBoxMaxY = max(lt[1], rt[1], lb[1], rb[1])
        boundingBox = (
            (boundingBoxMinX, boundingBoxMinY),
            (boundingBoxMaxX, boundingBoxMaxY))
        points = {}
        xrange = list(range(boundingBox[0][0], boundingBox[1][0]))
        yrange = list(range(boundingBox[0][1], boundingBox[1][1]))
        transformedX = float(centre[0])
        transformedY = float(centre[1])
        cx = float(centre[0])
        cy = float(centre[1])
        for xx in xrange:
            for yy in yrange:
                newX = xx * math.cos(angle) + yy * math.sin(angle)
                newY = -xx * math.sin(angle) + yy * math.cos(angle)

                if (newX - transformedX < width
                        and newY - transformedY < height
                        and newX - transformedX > 0
                        and newY - transformedY > 0):
                    points[(int(x + cx), int(y + cy))] = 1
        return points

##
# The workflow data object, which wraps the omero.mdoel.NamespaceI class
#


class WorkflowData(DataObject):

    def __init__(self, workflow=None):
        DataObject.__init__(self)
        if(workflow is None):
            self.setValue(NamespaceI())
            self.setNamespace("")
            self.setKeywords([])
            self.setDirty(True)
        else:
            self.setValue(workflow)

    ##
    # Set the namespace of the workflow.
    # @param namespace See above.
    def setNamespace(self, namespace):
        workflow = self.asIObject()
        if(workflow is None):
            raise Exception("No workflow specified.")
        workflow.setName(rstring(namespace))
        self.setDirty(True)

    ##
    # Get the namespace of the workflow
    # @return See Above.
    def getNamespace(self):
        workflow = self.asIObject()
        if(workflow is None):
            raise Exception("No Workflow specified.")
        namespace = workflow.getName()
        if(namespace is None):
            return ""
        return namespace.getValue()

    ##
    # Set the keywords of the workflow.
    # @param namespace See above.
    def setKeywords(self, keywords):
        workflow = self.asIObject()
        if(workflow is None):
            raise Exception("No workflow specified.")
        workflow.setKeywords(keywords)
        self.setDirty(True)

    ##
    # Set the keywords of the workflow.
    # @param namespace See above.
    def setKeywordsFromString(self, keywords):
        workflow = self.asIObject()
        if(workflow is None):
            raise Exception("No workflow specified.")
        workflow.setKeywords(toList(keywords))
        self.setDirty(True)

    ##
    # Get the keywords of the workflow
    # @return See Above.
    def getKeywords(self):
        workflow = self.asIObject()
        if(workflow is None):
            raise Exception("No Workflow specified.")
        keywords = workflow.getKeywords()
        if(keywords is None):
            return []
        return keywords

    ##
    # Add a keyword to the workflow
    # @param keyword See Above.
    def addKeyword(self, keyword):
        if(self.containsKeyword(keyword)):
            return
        keywords = self.getKeywords()
        keywords.append(keyword)
        self.setKeywords(keywords)

    ##
    # Return <code>True</code> if the keyword is part of workflow
    # @return See Above.
    def containsKeyword(self, keyword):
        keywords = self.getKeywords()
        return (keyword in keywords)

    ##
    # Remove the keyword from the workflow
    # @param keyword See Above.
    def removeKeyword(self, keyword):
        if(not self.containsKeyword(keyword)):
            return
        newList = self.getKeywords()
        newList.remove(keyword)
        self.setKeywords(newList)



#!/usr/bin/env python

#
# :author: Josh Moore <josh@glencoesoftware.com>
#
# Copyright (c) 2007, Glencoe Software, Inc.
# See LICENSE for details.
#

"""
Python driver for OMERO
"""
from __future__ import print_function

import logging
import os
import sys

import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)
# Assuming a CLI user never wants to see development details
# such as code that has been deprecated.


def not_root():
    """
    Check that the effective current user is not 0
    on systems supporting os.geteuid()
    """
    try:
        euid = os.geteuid()
        if euid == 0:
            print(
                "FATAL: Running %s as root can corrupt your directory "
                "permissions." % sys.argv[0])
            sys.exit(1)
        else:
            return euid
    except AttributeError:
        # This platform doesn't support effective uid
        # So nothing to worry about.
        return None


def readlink(file=sys.argv[0]):
    """
    Resolve symlinks and similar. This is useful to allow
    linking this file under /usr/bin/, for example.
    """
    import stat

    file = sys.argv[0]
    if not os.path.exists(file) and sys.platform == 'win32':
        file += '.exe'
    while stat.S_ISLNK(os.lstat(file)[stat.ST_MODE]):
        target = os.readlink(file)
        if target[0] != "/":
            file = os.path.join(os.path.dirname(file), target)
        else:
            file = target

    file = os.path.abspath(file)
    return file


def main():
    not_root()

    if "OMERO_HOME" in os.environ:
        print("WARN: OMERO_HOME usage is ignored in omero-py")

    exe = readlink()
    top = os.path.join(exe, os.pardir, os.pardir)

    #
    # This list needs to be kept in line with omero.cli.CLI._env
    #
    top = os.path.normpath(top)
    var = os.path.join(top, "var")
    vlb = os.path.join(var, "lib")
    sys.path.append(vlb)

    # Testing shortcut. If the first argument is an
    # empty string, exit sucessfully.
    #
    if len(sys.argv) == 2 and sys.argv[1] == "":
        sys.exit(0)

    #
    # Primary activity: import omero.cli and launch
    # catching any Ctrl-Cs from the user
    #
    try:
        try:
            import omero.cli
        except ImportError as ie:
            OMERODIR = os.environ.get('OMERODIR', None)
            print("*"*80)
            print("""
            ERROR: Could not import omero.cli! (%s)

            This means that your installation is incomplete. Contact
            the OME mailing lists for more information:

            https://www.openmicroscopy.org/support/

            If you are building from source, please supply the build log
            as well as which version you are building from. If you
            downloaded a distribution, please provide which link you
            used.
            """ % ie)
            print("*"*80)
            print("""
            Debugging Info:
            --------------
            CWD=%s
            VERSION=%s
            OMERO_EXE=%s
            OMERODIR=%s
            PYTHONPATH=%s
            """ % (os.getcwd(), sys.version.replace("\n", " "), top,
                   OMERODIR, sys.path))
            sys.exit(2)

        logging.basicConfig(level=logging.WARN)
        rv = omero.cli.argv()
        sys.exit(rv)
    except KeyboardInterrupt:
        print("Cancelled")
        sys.exit(1)



# -*- coding: utf-8 -*-
# **********************************************************************
#
# Copyright (c) 2003-2017 ZeroC, Inc. All rights reserved.
#
# This copy of Ice is licensed to you under the terms described in the
# ICE_LICENSE file included in this distribution.
#
# **********************************************************************
#
# Ice version 3.6.4
#
# <auto-generated>
#
# Generated from file `ITimeline.ice'
#
# Warning: do not edit this file.
#
# </auto-generated>
#

from sys import version_info as _version_info_
import Ice, IcePy
import omero_ModelF_ice
import omero_ServicesF_ice
import omero_System_ice
import omero_Collections_ice

# Included module omero
_M_omero = Ice.openModule('omero')

# Included module omero.model
_M_omero.model = Ice.openModule('omero.model')

# Included module Ice
_M_Ice = Ice.openModule('Ice')

# Included module Glacier2
_M_Glacier2 = Ice.openModule('Glacier2')

# Included module omero.sys
_M_omero.sys = Ice.openModule('omero.sys')

# Included module omero.api
_M_omero.api = Ice.openModule('omero.api')

# Included module omero.grid
_M_omero.grid = Ice.openModule('omero.grid')

# Start of module omero
__name__ = 'omero'

# Start of module omero.api
__name__ = 'omero.api'

if 'ITimeline' not in _M_omero.api.__dict__:
    _M_omero.api.ITimeline = Ice.createTempClass()
    class ITimeline(_M_omero.api.ServiceInterface):
        """
        Service for the querying of OMERO metadata based on creation and
        modification time. Currently supported types for querying include:
        omero.model.Annotation
        omero.model.Dataset
        omero.model.Image
        omero.model.Project
        omero.model.RenderingDef
        Return maps
        The map return values will be indexed by the short type name above:
        Project, Image, ... All keys which are passed in the
        StringSet argument will be included in the returned map, even if
        they have no values. A default value of 0 or the empty list \[]
        will be used.
        The only exception to this rule is that the null or empty StringSet
        implies all valid keys.
        Parameters
        All methods take a omero::sys::Parameters object and will apply the filter
        object for paging through the data in order to prevent loading too
        many objects. If the parameters argument is null or no paging is activated,
        then the default will be: OFFSET=0, LIMIT=50. Filter.ownerId and
        Filter.groupId will also be AND'ed to the query if either value is present.
        If both are null, then the current user id will be used. To retrieve for
        all users, use ownerId == rlong(-1) and
        groupId == null.
        Merging
        The methods which take a StringSet and a Parameters object, also
        have a bool merge argument. This argument defines whether or
        not the LIMIT applies to each object independently
        (\["a","b"] @ 100 == 200) or merges the lists together
        chronologically (\["a","b"] @ 100 merged == 100).
        Time used
        Except for Image, IObject.details.updateEvent is used in all cases
        for determining most recent events. Images are compared via
        Image.acquisitionDate which is unlike the other properties is
        modifiable by the user.
        A typical invocation might look like (in Python):
        {@code
        timeline = sf.getTimelineService()
        params = ParametersI().page(0,100)
        types = \["Project","Dataset"])
        map = timeline.getByPeriod(types, params, False)
        }
        At this point, map will not contain more than 200 objects.
        This service is defined only in Blitz and so no javadoc is available
        in the ome.api package.
        TODOS: binning, stateful caching, ...
        """
        def __init__(self):
            if Ice.getType(self) == _M_omero.api.ITimeline:
                raise RuntimeError('omero.api.ITimeline is an abstract class')

        def ice_ids(self, current=None):
            return ('::Ice::Object', '::omero::api::ITimeline', '::omero::api::ServiceInterface')

        def ice_id(self, current=None):
            return '::omero::api::ITimeline'

        def ice_staticId():
            return '::omero::api::ITimeline'
        ice_staticId = staticmethod(ice_staticId)

        def getMostRecentAnnotationLinks_async(self, _cb, parentTypes, childTypes, namespaces, p, current=None):
            """
            Return the last LIMIT annotation __Links__ whose parent (IAnnotated)
            matches one of the parentTypes, whose child (Annotation) matches one
            of the childTypes (limit of one for the moment), and who namespace
            matches via ILIKE.
            The parents and children will be unloaded. The link will have
            its creation/update events loaded.
            If Parameters.theFilter.ownerId or groupId are set, they will be
            AND'ed to the query to filter results.
            Merges by default based on parentType.
            Arguments:
            _cb -- The asynchronous callback object.
            parentTypes -- 
            childTypes -- 
            namespaces -- 
            p -- 
            current -- The Current object for the invocation.
            """
            pass

        def getMostRecentShareCommentLinks_async(self, _cb, p, current=None):
            """
            Return the last LIMIT comment annotation links attached to a share by
            __others__.
            Note: Currently the storage of these objects is not optimal
            and so this method may change.
            Arguments:
            _cb -- The asynchronous callback object.
            p -- 
            current -- The Current object for the invocation.
            """
            pass

        def getMostRecentObjects_async(self, _cb, types, p, merge, current=None):
            """
            Returns the last LIMIT objects of TYPES as ordered by
            creation/modification times in the Event table.
            Arguments:
            _cb -- The asynchronous callback object.
            types -- 
            p -- 
            merge -- 
            current -- The Current object for the invocation.
            """
            pass

        def getByPeriod_async(self, _cb, types, start, end, p, merge, current=None):
            """
            Returns the given LIMIT objects of TYPES as ordered by
            creation/modification times in the Event table, but
            within the given time window.
            Arguments:
            _cb -- The asynchronous callback object.
            types -- 
            start -- 
            end -- 
            p -- 
            merge -- 
            current -- The Current object for the invocation.
            """
            pass

        def countByPeriod_async(self, _cb, types, start, end, p, current=None):
            """
            Queries the same information as getByPeriod, but only returns the counts
            for the given objects.
            Arguments:
            _cb -- The asynchronous callback object.
            types -- 
            start -- 
            end -- 
            p -- 
            current -- The Current object for the invocation.
            """
            pass

        def getEventLogsByPeriod_async(self, _cb, start, end, p, current=None):
            """
            Returns the EventLog table objects which are queried to produce the counts above.
            Note the concept of period inclusion mentioned above.
            WORKAROUND WARNING: this method returns non-managed EventLogs (i.e.
            eventLog.getId() == null) for Image acquisitions.
            Arguments:
            _cb -- The asynchronous callback object.
            start -- 
            end -- 
            p -- 
            current -- The Current object for the invocation.
            """
            pass

        def __str__(self):
            return IcePy.stringify(self, _M_omero.api._t_ITimeline)

        __repr__ = __str__

    _M_omero.api.ITimelinePrx = Ice.createTempClass()
    class ITimelinePrx(_M_omero.api.ServiceInterfacePrx):

        """
        Return the last LIMIT annotation __Links__ whose parent (IAnnotated)
        matches one of the parentTypes, whose child (Annotation) matches one
        of the childTypes (limit of one for the moment), and who namespace
        matches via ILIKE.
        The parents and children will be unloaded. The link will have
        its creation/update events loaded.
        If Parameters.theFilter.ownerId or groupId are set, they will be
        AND'ed to the query to filter results.
        Merges by default based on parentType.
        Arguments:
        parentTypes -- 
        childTypes -- 
        namespaces -- 
        p -- 
        _ctx -- The request context for the invocation.
        """
        def getMostRecentAnnotationLinks(self, parentTypes, childTypes, namespaces, p, _ctx=None):
            return _M_omero.api.ITimeline._op_getMostRecentAnnotationLinks.invoke(self, ((parentTypes, childTypes, namespaces, p), _ctx))

        """
        Return the last LIMIT annotation __Links__ whose parent (IAnnotated)
        matches one of the parentTypes, whose child (Annotation) matches one
        of the childTypes (limit of one for the moment), and who namespace
        matches via ILIKE.
        The parents and children will be unloaded. The link will have
        its creation/update events loaded.
        If Parameters.theFilter.ownerId or groupId are set, they will be
        AND'ed to the query to filter results.
        Merges by default based on parentType.
        Arguments:
        parentTypes -- 
        childTypes -- 
        namespaces -- 
        p -- 
        _response -- The asynchronous response callback.
        _ex -- The asynchronous exception callback.
        _sent -- The asynchronous sent callback.
        _ctx -- The request context for the invocation.
        Returns: An asynchronous result object for the invocation.
        """
        def begin_getMostRecentAnnotationLinks(self, parentTypes, childTypes, namespaces, p, _response=None, _ex=None, _sent=None, _ctx=None):
            return _M_omero.api.ITimeline._op_getMostRecentAnnotationLinks.begin(self, ((parentTypes, childTypes, namespaces, p), _response, _ex, _sent, _ctx))

        """
        Return the last LIMIT annotation __Links__ whose parent (IAnnotated)
        matches one of the parentTypes, whose child (Annotation) matches one
        of the childTypes (limit of one for the moment), and who namespace
        matches via ILIKE.
        The parents and children will be unloaded. The link will have
        its creation/update events loaded.
        If Parameters.theFilter.ownerId or groupId are set, they will be
        AND'ed to the query to filter results.
        Merges by default based on parentType.
        Arguments:
        parentTypes -- 
        childTypes -- 
        namespaces -- 
        p -- 
        """
        def end_getMostRecentAnnotationLinks(self, _r):
            return _M_omero.api.ITimeline._op_getMostRecentAnnotationLinks.end(self, _r)

        """
        Return the last LIMIT comment annotation links attached to a share by
        __others__.
        Note: Currently the storage of these objects is not optimal
        and so this method may change.
        Arguments:
        p -- 
        _ctx -- The request context for the invocation.
        """
        def getMostRecentShareCommentLinks(self, p, _ctx=None):
            return _M_omero.api.ITimeline._op_getMostRecentShareCommentLinks.invoke(self, ((p, ), _ctx))

        """
        Return the last LIMIT comment annotation links attached to a share by
        __others__.
        Note: Currently the storage of these objects is not optimal
        and so this method may change.
        Arguments:
        p -- 
        _response -- The asynchronous response callback.
        _ex -- The asynchronous exception callback.
        _sent -- The asynchronous sent callback.
        _ctx -- The request context for the invocation.
        Returns: An asynchronous result object for the invocation.
        """
        def begin_getMostRecentShareCommentLinks(self, p, _response=None, _ex=None, _sent=None, _ctx=None):
            return _M_omero.api.ITimeline._op_getMostRecentShareCommentLinks.begin(self, ((p, ), _response, _ex, _sent, _ctx))

        """
        Return the last LIMIT comment annotation links attached to a share by
        __others__.
        Note: Currently the storage of these objects is not optimal
        and so this method may change.
        Arguments:
        p -- 
        """
        def end_getMostRecentShareCommentLinks(self, _r):
            return _M_omero.api.ITimeline._op_getMostRecentShareCommentLinks.end(self, _r)

        """
        Returns the last LIMIT objects of TYPES as ordered by
        creation/modification times in the Event table.
        Arguments:
        types -- 
        p -- 
        merge -- 
        _ctx -- The request context for the invocation.
        """
        def getMostRecentObjects(self, types, p, merge, _ctx=None):
            return _M_omero.api.ITimeline._op_getMostRecentObjects.invoke(self, ((types, p, merge), _ctx))

        """
        Returns the last LIMIT objects of TYPES as ordered by
        creation/modification times in the Event table.
        Arguments:
        types -- 
        p -- 
        merge -- 
        _response -- The asynchronous response callback.
        _ex -- The asynchronous exception callback.
        _sent -- The asynchronous sent callback.
        _ctx -- The request context for the invocation.
        Returns: An asynchronous result object for the invocation.
        """
        def begin_getMostRecentObjects(self, types, p, merge, _response=None, _ex=None, _sent=None, _ctx=None):
            return _M_omero.api.ITimeline._op_getMostRecentObjects.begin(self, ((types, p, merge), _response, _ex, _sent, _ctx))

        """
        Returns the last LIMIT objects of TYPES as ordered by
        creation/modification times in the Event table.
        Arguments:
        types -- 
        p -- 
        merge -- 
        """
        def end_getMostRecentObjects(self, _r):
            return _M_omero.api.ITimeline._op_getMostRecentObjects.end(self, _r)

        """
        Returns the given LIMIT objects of TYPES as ordered by
        creation/modification times in the Event table, but
        within the given time window.
        Arguments:
        types -- 
        start -- 
        end -- 
        p -- 
        merge -- 
        _ctx -- The request context for the invocation.
        """
        def getByPeriod(self, types, start, end, p, merge, _ctx=None):
            return _M_omero.api.ITimeline._op_getByPeriod.invoke(self, ((types, start, end, p, merge), _ctx))

        """
        Returns the given LIMIT objects of TYPES as ordered by
        creation/modification times in the Event table, but
        within the given time window.
        Arguments:
        types -- 
        start -- 
        end -- 
        p -- 
        merge -- 
        _response -- The asynchronous response callback.
        _ex -- The asynchronous exception callback.
        _sent -- The asynchronous sent callback.
        _ctx -- The request context for the invocation.
        Returns: An asynchronous result object for the invocation.
        """
        def begin_getByPeriod(self, types, start, end, p, merge, _response=None, _ex=None, _sent=None, _ctx=None):
            return _M_omero.api.ITimeline._op_getByPeriod.begin(self, ((types, start, end, p, merge), _response, _ex, _sent, _ctx))

        """
        Returns the given LIMIT objects of TYPES as ordered by
        creation/modification times in the Event table, but
        within the given time window.
        Arguments:
        types -- 
        start -- 
        end -- 
        p -- 
        merge -- 
        """
        def end_getByPeriod(self, _r):
            return _M_omero.api.ITimeline._op_getByPeriod.end(self, _r)

        """
        Queries the same information as getByPeriod, but only returns the counts
        for the given objects.
        Arguments:
        types -- 
        start -- 
        end -- 
        p -- 
        _ctx -- The request context for the invocation.
        """
        def countByPeriod(self, types, start, end, p, _ctx=None):
            return _M_omero.api.ITimeline._op_countByPeriod.invoke(self, ((types, start, end, p), _ctx))

        """
        Queries the same information as getByPeriod, but only returns the counts
        for the given objects.
        Arguments:
        types -- 
        start -- 
        end -- 
        p -- 
        _response -- The asynchronous response callback.
        _ex -- The asynchronous exception callback.
        _sent -- The asynchronous sent callback.
        _ctx -- The request context for the invocation.
        Returns: An asynchronous result object for the invocation.
        """
        def begin_countByPeriod(self, types, start, end, p, _response=None, _ex=None, _sent=None, _ctx=None):
            return _M_omero.api.ITimeline._op_countByPeriod.begin(self, ((types, start, end, p), _response, _ex, _sent, _ctx))

        """
        Queries the same information as getByPeriod, but only returns the counts
        for the given objects.
        Arguments:
        types -- 
        start -- 
        end -- 
        p -- 
        """
        def end_countByPeriod(self, _r):
            return _M_omero.api.ITimeline._op_countByPeriod.end(self, _r)

        """
        Returns the EventLog table objects which are queried to produce the counts above.
        Note the concept of period inclusion mentioned above.
        WORKAROUND WARNING: this method returns non-managed EventLogs (i.e.
        eventLog.getId() == null) for Image acquisitions.
        Arguments:
        start -- 
        end -- 
        p -- 
        _ctx -- The request context for the invocation.
        """
        def getEventLogsByPeriod(self, start, end, p, _ctx=None):
            return _M_omero.api.ITimeline._op_getEventLogsByPeriod.invoke(self, ((start, end, p), _ctx))

        """
        Returns the EventLog table objects which are queried to produce the counts above.
        Note the concept of period inclusion mentioned above.
        WORKAROUND WARNING: this method returns non-managed EventLogs (i.e.
        eventLog.getId() == null) for Image acquisitions.
        Arguments:
        start -- 
        end -- 
        p -- 
        _response -- The asynchronous response callback.
        _ex -- The asynchronous exception callback.
        _sent -- The asynchronous sent callback.
        _ctx -- The request context for the invocation.
        Returns: An asynchronous result object for the invocation.
        """
        def begin_getEventLogsByPeriod(self, start, end, p, _response=None, _ex=None, _sent=None, _ctx=None):
            return _M_omero.api.ITimeline._op_getEventLogsByPeriod.begin(self, ((start, end, p), _response, _ex, _sent, _ctx))

        """
        Returns the EventLog table objects which are queried to produce the counts above.
        Note the concept of period inclusion mentioned above.
        WORKAROUND WARNING: this method returns non-managed EventLogs (i.e.
        eventLog.getId() == null) for Image acquisitions.
        Arguments:
        start -- 
        end -- 
        p -- 
        """
        def end_getEventLogsByPeriod(self, _r):
            return _M_omero.api.ITimeline._op_getEventLogsByPeriod.end(self, _r)

        def checkedCast(proxy, facetOrCtx=None, _ctx=None):
            return _M_omero.api.ITimelinePrx.ice_checkedCast(proxy, '::omero::api::ITimeline', facetOrCtx, _ctx)
        checkedCast = staticmethod(checkedCast)

        def uncheckedCast(proxy, facet=None):
            return _M_omero.api.ITimelinePrx.ice_uncheckedCast(proxy, facet)
        uncheckedCast = staticmethod(uncheckedCast)

        def ice_staticId():
            return '::omero::api::ITimeline'
        ice_staticId = staticmethod(ice_staticId)

    _M_omero.api._t_ITimelinePrx = IcePy.defineProxy('::omero::api::ITimeline', ITimelinePrx)

    _M_omero.api._t_ITimeline = IcePy.defineClass('::omero::api::ITimeline', ITimeline, -1, (), True, False, None, (_M_omero.api._t_ServiceInterface,), ())
    ITimeline._ice_type = _M_omero.api._t_ITimeline

    ITimeline._op_getMostRecentAnnotationLinks = IcePy.Operation('getMostRecentAnnotationLinks', Ice.OperationMode.Idempotent, Ice.OperationMode.Idempotent, True, None, (), (((), _M_omero.api._t_StringSet, False, 0), ((), _M_omero.api._t_StringSet, False, 0), ((), _M_omero.api._t_StringSet, False, 0), ((), _M_omero.sys._t_Parameters, False, 0)), (), ((), _M_omero.api._t_IObjectList, False, 0), (_M_omero._t_ServerError,))
    ITimeline._op_getMostRecentShareCommentLinks = IcePy.Operation('getMostRecentShareCommentLinks', Ice.OperationMode.Idempotent, Ice.OperationMode.Idempotent, True, None, (), (((), _M_omero.sys._t_Parameters, False, 0),), (), ((), _M_omero.api._t_IObjectList, False, 0), (_M_omero._t_ServerError,))
    ITimeline._op_getMostRecentObjects = IcePy.Operation('getMostRecentObjects', Ice.OperationMode.Idempotent, Ice.OperationMode.Idempotent, True, None, (), (((), _M_omero.api._t_StringSet, False, 0), ((), _M_omero.sys._t_Parameters, False, 0), ((), IcePy._t_bool, False, 0)), (), ((), _M_omero.api._t_IObjectListMap, False, 0), (_M_omero._t_ServerError,))
    ITimeline._op_getByPeriod = IcePy.Operation('getByPeriod', Ice.OperationMode.Idempotent, Ice.OperationMode.Idempotent, True, None, (), (((), _M_omero.api._t_StringSet, False, 0), ((), _M_omero._t_RTime, False, 0), ((), _M_omero._t_RTime, False, 0), ((), _M_omero.sys._t_Parameters, False, 0), ((), IcePy._t_bool, False, 0)), (), ((), _M_omero.api._t_IObjectListMap, False, 0), (_M_omero._t_ServerError,))
    ITimeline._op_countByPeriod = IcePy.Operation('countByPeriod', Ice.OperationMode.Idempotent, Ice.OperationMode.Idempotent, True, None, (), (((), _M_omero.api._t_StringSet, False, 0), ((), _M_omero._t_RTime, False, 0), ((), _M_omero._t_RTime, False, 0), ((), _M_omero.sys._t_Parameters, False, 0)), (), ((), _M_omero.api._t_StringLongMap, False, 0), (_M_omero._t_ServerError,))
    ITimeline._op_getEventLogsByPeriod = IcePy.Operation('getEventLogsByPeriod', Ice.OperationMode.Idempotent, Ice.OperationMode.Idempotent, True, None, (), (((), _M_omero._t_RTime, False, 0), ((), _M_omero._t_RTime, False, 0), ((), _M_omero.sys._t_Parameters, False, 0)), (), ((), _M_omero.api._t_EventLogList, False, 0), (_M_omero._t_ServerError,))

    _M_omero.api.ITimeline = ITimeline
    del ITimeline

    _M_omero.api.ITimelinePrx = ITimelinePrx
    del ITimelinePrx

# End of module omero.api

__name__ = 'omero'

# End of module omero



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017 University of Dundee & Open Microscopy Environment.
# All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
Utility methods for dealing with scripts.
"""
from __future__ import division

from future import standard_library
standard_library.install_aliases()
from past.utils import old_div
try:
    from PIL import Image, ImageDraw, ImageFont  # see ticket:2597
except ImportError:
    import Image
    import ImageDraw
    import ImageFont  # see ticket:2597

import os.path
import omero.gateway
import io
from omero.rtypes import rint

GATEWAYPATH = omero.gateway.THISPATH


def get_font(fontsize):
    """
    Returns a PIL ImageFont Sans-serif true-type font of the specified size
    or a pre-compiled font of fixed size if the ttf font is not found

    :param fontsize:	The size of the font you want
    :return: 	A PIL Font
    """

    font_path = os.path.join(GATEWAYPATH, "pilfonts", "FreeSans.ttf")
    try:
        font = ImageFont.truetype(font_path, fontsize)
    except:
        font = ImageFont.load('%s/pilfonts/B%0.2d.pil' % (GATEWAYPATH, 24))
    return font


def paste_image(image, canvas, x, y):
    """
    Pastes the image onto the canvas at the specified coordinates
    Image and canvas are instances of PIL 'Image'

    :param image:		The PIL image to be pasted. Image
    :param canvas:		The PIL image on which to paste. Image
    :param x:			X coordinate (left) to paste
    :param y: 			Y coordinate (top) to paste
    """

    x_right = image.size[0] + x
    y_bottom = image.size[1] + y
    # make a tuple of topleft-x, topleft-y, bottomRight-x, bottomRight-y
    pastebox = (x, y, x_right, y_bottom)
    canvas.paste(image, pastebox)


def paint_thumbnail_grid(thumbnail_store, length, spacing, pixel_ids,
                         col_count, bg=(255, 255, 255), left_label=None,
                         text_color=(0, 0, 0), fontsize=None, top_label=None):
    """
    Retrieves thumbnails for each pixelId, and places them in a grid,
    with White background.
    Option to add a vertical label to the left of the canvas
    Creates a PIL 'Image' which is returned

    :param thumbnail_store:  The omero thumbnail store.
    :param length:			 Length of longest thumbnail side, int
    :param spacing:			 The spacing between thumbnails and around the
                             edges. int
    :param pixel_ids:		 List of pixel IDs. [long]
    :param col_count:		 The number of columns. int
    :param bg:				 Background colour as (r,g,b).
                             Default is white (255, 255, 255)
    :param left_label: 		 Optional string to display vertically to the left.
    :param text_color:		 The color of the text as (r,g,b).
                             Default is black (0, 0, 0)
    :param fontsize:		 Size of the font.
                             Default is calculated based on thumbnail length,
                             int
    :return: 			    The PIL Image canvas.
    """
    mode = "RGB"
    # work out how many rows and columns are needed for all the images
    img_count = len(pixel_ids)

    row_count = (old_div(img_count, col_count))
    # check that we have enough rows and cols...
    while (col_count * row_count) < img_count:
        row_count += 1

    left_space = top_space = spacing
    min_width = 0

    text_height = 0
    if left_label or top_label:
        # if no images (no rows), need to make at least one row to show label
        if left_label is not None and row_count == 0:
            row_count = 1
        if fontsize is None:
            fontsize = old_div(length, 10) + 5
        font = get_font(fontsize)
        if left_label:
            text_width, text_height = font.getsize(left_label)
            left_space = spacing + text_height + spacing
        if top_label:
            text_width, text_height = font.getsize(top_label)
            top_space = spacing + text_height + spacing
            min_width = left_space + text_width + spacing

    # work out the canvas size needed, and create a white canvas
    cols_needed = min(col_count, img_count)
    v = left_space + cols_needed * (length + spacing)
    canvas_width = max(min_width, v)
    canvas_height = top_space + row_count * (length + spacing) + spacing
    mode = "RGB"
    size = (canvas_width, canvas_height)
    canvas = Image.new(mode, size, bg)

    # to write text up the left side, need to write it on horizontal canvas
    # and rotate.
    if left_label:
        label_canvas_width = canvas_height
        label_canvas_height = text_height + spacing
        label_size = (label_canvas_width, label_canvas_height)
        text_canvas = Image.new(mode, label_size, bg)
        draw = ImageDraw.Draw(text_canvas)
        text_width = font.getsize(left_label)[0]
        text_x = old_div((label_canvas_width - text_width), 2)
        draw.text((text_x, spacing), left_label, font=font, fill=text_color)
        vertical_canvas = text_canvas.rotate(90)
        paste_image(vertical_canvas, canvas, 0, 0)
        del draw

    if top_label is not None:
        label_canvas_width = canvas_width
        label_canvas_height = text_height + spacing
        label_size = (label_canvas_width, label_canvas_height)
        text_canvas = Image.new(mode, label_size, bg)
        draw = ImageDraw.Draw(text_canvas)
        draw.text((spacing, spacing), top_label, font=font, fill=text_color)
        paste_image(text_canvas, canvas, left_space, 0)
        del draw

    # loop through the images, getting a thumbnail and placing it on a new row
    # and column
    r = 0
    c = 0
    thumbnail_map = thumbnail_store.getThumbnailByLongestSideSet(rint(length),
                                                                 pixel_ids)
    for pixels_id in pixel_ids:
        if pixels_id in thumbnail_map:
            thumbnail = thumbnail_map[pixels_id]
            # check we have a thumbnail (won't get one if image is invalid)
            if thumbnail:
                # make an "Image" from the string-encoded thumbnail
                thumb_image = Image.open(io.StringIO(thumbnail))
                # paste the image onto the canvas at the correct coordinates
                # for the current row and column
                x = c * (length + spacing) + left_space
                y = r * (length + spacing) + top_space
                paste_image(thumb_image, canvas, x, y)

        # increment the column, and if we're at the last column, start a new
        # row
        c = c + 1
        if c == col_count:
            c = 0
            r = r + 1

    return canvas


def int_to_rgba(rgba):
    """
    Returns a tuple of (r,g,b,a) from an integer color
    r, g, b, a are 0-255.

    :param rgba:		A color as integer. Int
    :return:		A tuple of (r,g,b,a)
    """
    a = check_rgb_range(rgba % 256)
    b = check_rgb_range(rgba / 256 % 256)
    g = check_rgb_range(rgba / 256 / 256 % 256)
    r = check_rgb_range(rgba / 256 / 256 / 256 % 256)
    if a == 0:
        a = 255
    return (r, g, b, a)


def check_rgb_range(value):
    """
    Checks that the value is between 0 and 255. Returns integer value
    If the value is not valid, return 255
    (better to see something than nothing!)

    :param value:		The value to check.
    :return:			An integer between 0 and 255
    """
    try:
        v = int(value)
        if 0 <= v <= 255:
            return v
    except:
        return 255


def get_zoom_factor(image_size, max_width, max_height):
    """
    Returns the factor by which the Image has to be shrunk
    so that its dimensions are less that max_width and max_height
    e.g. if the image must be half-sized, this method returns 2.0 (float)

    :param image_size:  Size of the image as tuple (width, height)
    :param max_width:   The max width after zooming
    :param max_height:  The max height after zooming
    :return: The factor by which to shrink the image to be
             within max width and height
    """
    image_width, imageheight = image_size
    zoom_width = old_div(float(image_width), float(max_width))
    zoom_height = old_div(float(imageheight), float(max_height))
    return max(zoom_width, zoom_height)


def resize_image(image, max_width, max_height):
    """
    Resize the image so that it is as big as possible,
    within the dimensions max_width, max_height

    :param image: The PIL Image to zoom
    :param max_width: The max width of the zoomed image
    :param max_height: The max height of the zoomed image
    :return: The zoomed image. PIL Image.
    """
    image_width, image_height = image.size
    if image_width == max_width and image_height == max_height:
        return image
    # find which axis requires the biggest zoom (smallest relative max
    # dimension)
    zoom_width = old_div(float(image_width), float(max_width))
    zoom_height = old_div(float(image_height), float(max_height))
    zoom = max(zoom_width, zoom_height)
    if zoom_width >= zoom_height:  # size is defined by width
        max_height = int(image_height // zoom)  # calculate the new height
    else:
        max_width = int(image_width // zoom)
    return image.resize((max_width, max_height))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# Copyright (C) 2014-2016 Glencoe Software, Inc. All Rights Reserved.
# Use is subject to license terms supplied in LICENSE.txt
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
fs plugin for querying repositories, filesets, and the like.
"""
from __future__ import division
from __future__ import print_function

from past.builtins import cmp
from builtins import zip
from builtins import str
from builtins import map
from builtins import object
from past.utils import old_div
import platform
import sys

from collections import defaultdict
from collections import namedtuple

from omero import client as Client
from omero import CmdError
from omero import ResourceError
from omero import ServerError
from omero import ValidationException
from omero.cli import admin_only
from omero.cli import CmdControl
from omero.cli import CLI
from omero.cli import ProxyStringType
from omero.gateway import BlitzGateway
from omero.model.enums import (
    AdminPrivilegeChown,
    AdminPrivilegeWriteOwned, AdminPrivilegeWriteManagedRepo,
    AdminPrivilegeDeleteOwned, AdminPrivilegeDeleteManagedRepo)
from omero.rtypes import rstring
from omero.rtypes import unwrap
from omero.sys import Principal
from omero.util.temp_files import create_path
from omero.util.text import filesizeformat
from omero.fs import TRANSFERS

from omero.install.windows_warning import windows_warning, WINDOWS_WARNING

HELP = """Filesystem utilities"""


if platform.system() == 'Windows':
    HELP += ("\n\n%s" % WINDOWS_WARNING)

Entry = namedtuple("Entry", ("level", "id", "path", "mimetype"))


def contents(mrepo, path, ctx=None):
    """
    Yield Entry namedtuples for each return value
    from treeList for the given path.
    """
    tree = unwrap(mrepo.treeList(path, ctx))

    def parse(tree, level=0):
        for k, v in list(tree.items()):
            yield Entry(level, v.get("id"),
                        k, v.get("mimetype"))
            if "files" in v:
                for sub in parse(v.get("files"), level+1):
                    yield sub

    for entry in parse(tree):
        yield entry


def prep_directory(client, mrepo):
    """
    Create an empty FS directory by performing an import and
    then deleting the created fileset.
    """

    from omero.cmd import Delete2, DoAll
    from omero.grid import ImportSettings

    from omero.model import ChecksumAlgorithmI
    from omero.model import FilesetI
    from omero.model import FilesetEntryI
    from omero.model import UploadJobI

    fs = FilesetI()
    fs.linkJob(UploadJobI())
    entry = FilesetEntryI()
    entry.clientPath = rstring("README.txt")
    fs.addFilesetEntry(entry)
    settings = ImportSettings()
    settings.checksumAlgorithm = ChecksumAlgorithmI()
    settings.checksumAlgorithm.value = rstring("SHA1-160")
    proc = mrepo.importFileset(fs, settings)
    try:

        tmp = create_path()
        prx = proc.getUploader(0)
        try:
            tmp.write_text("THIS IS A PLACEHOLDER")
            hash = client.sha1(tmp)
            with open(tmp, "rb") as source:
                client.write_stream(source, prx)
        finally:
            prx.close()
        tmp.remove()

        handle = proc.verifyUpload([hash])
        try:
            req = handle.getRequest()
            fs = req.activity.parent
        finally:
            handle.close()

        dir = unwrap(mrepo.treeList(fs.templatePrefix.val))
        oid = list(dir.items())[0][1].get("id")
        ofile = client.sf.getQueryService().get("OriginalFile", oid)

        delete1 = Delete2(targetObjects={'Fileset': [fs.id.val]})
        delete2 = Delete2(targetObjects={'OriginalFile': [ofile.id.val]})
        doall = DoAll()
        doall.requests = [delete1, delete2]
        cb = client.submit(doall)
        cb.close(True)

    finally:
        proc.close()

    return fs.templatePrefix.val


def get_logfile(query, fid):
    from omero.sys import ParametersI
    q = ("select o from FilesetJobLink l "
         "join l.parent as fs join l.child as j "
         "join j.originalFileLinks l2 join l2.child as o "
         "where fs.id = :id and "
         "o.mimetype = 'application/omero-log-file'")
    return query.findByQuery(q, ParametersI().addId(fid))


def rename_fileset(client, mrepo, fileset, new_dir, ctx=None):
    """
    Loads each OriginalFile found under orig_dir and
    updates its path field to point at new_dir. Files
    are not yet moved.
    """

    from omero.constants.namespaces import NSFSRENAME
    from omero.model import CommentAnnotationI
    from omero.model import FilesetAnnotationLinkI

    tomove = []
    tosave = []
    query = client.sf.getQueryService()
    update = client.sf.getUpdateService()
    orig_dir = fileset.templatePrefix.val

    def parse_parent(dir):
        """
        Note that final elements are empty
        """
        parts = dir.split("/")
        parpath = "/".join(parts[0:-2]+[""])
        parname = parts[-2]
        logname = parts[-2] + ".log"
        return parpath, parname, logname

    orig_parpath, orig_parname, orig_logname = parse_parent(orig_dir)
    new_parpath, new_parname, new_logname = parse_parent(new_dir)

    for entry in contents(mrepo, orig_dir, ctx):

        ofile = query.get("OriginalFile", entry.id, ctx)
        path = ofile.path.val

        if entry.level == 0:
            tomove.append((orig_dir, new_dir))
            assert orig_parpath in path
            repl = path.replace(orig_parpath, new_parpath)
            ofile.name = rstring(new_parname)
        else:
            assert orig_dir in path
            repl = path.replace(orig_dir, new_dir)
        ofile.path = rstring(repl)
        tosave.append(ofile)
    fileset.templatePrefix = rstring(new_dir)
    # TODO: placing the fileset at the end of this list
    # causes ONLY the fileset to be updated !!
    tosave.insert(0, fileset)

    # Add an annotation to the fileset as well so
    # we can detect if something's gone wrong.
    link = FilesetAnnotationLinkI()
    link.parent = fileset.proxy()
    link.child = CommentAnnotationI()
    link.child.ns = rstring(NSFSRENAME)
    link.child.textValue = rstring("previous=%s" % orig_dir)
    tosave.insert(1, link)

    # And now move the log file as well:
    log = get_logfile(query, fileset.id.val)
    if log is not None:
        target = new_parpath + new_logname
        source = orig_parpath + orig_logname
        tomove.append((source, target))
        log.path = rstring(new_parpath)
        log.name = rstring(new_logname)
        tosave.append(log)

    # Done. Save in one transaction and return tomove
    update.saveAndReturnArray(tosave, ctx)
    return tomove


class FsControl(CmdControl):

    def _configure(self, parser):

        parser.add_login_arguments()
        sub = parser.sub()

        images = parser.add(sub, self.images)
        images.add_style_argument()
        images.add_limit_arguments()
        images.add_argument(
            "--order", default="newest",
            choices=("newest", "oldest", "largest"),
            help="order of the rows returned")
        images.add_argument(
            "--archived", action="store_true",
            help="list only images with archived data")

        mkdir = parser.add(sub, self.mkdir)
        mkdir.add_argument(
            "new_dir",
            help="directory to create in the repository")
        mkdir.add_argument(
            "--parents", action="store_true",
            help="ensure whole path exists")

        rename = parser.add(sub, self.rename)
        rename.add_argument(
            "fileset",
            type=ProxyStringType("Fileset"),
            help="Fileset which should be renamed: ID or Fileset:ID")
        rename.add_argument(
            "--no-move", action="store_true",
            help="do not move original files and import log")

        repos = parser.add(sub, self.repos)
        repos.add_style_argument()
        repos.add_argument(
            "--managed", action="store_true",
            help="repos only managed repositories")

        sets = parser.add(sub, self.sets)
        sets.add_style_argument()
        sets.add_limit_arguments()
        sets.add_argument(
            "--order", default="newest",
            choices=("newest", "oldest", "prefix"),
            help="order of the rows returned")
        sets.add_argument(
            "--without-images", action="store_true",
            help="list only sets without images (i.e. corrupt)")
        sets.add_argument(
            "--with-transfer", nargs="+", action="append",
            help="list sets by their in-place import method")
        sets.add_argument(
            "--check", action="store_true",
            help="verify the file checksums for each fileset (admins only)")

        ls = parser.add(sub, self.ls)
        ls.add_argument(
            "fileset",
            type=ProxyStringType("Fileset"))

        logfile = parser.add(sub, self.logfile)
        logfile.add_argument("fileset", type=ProxyStringType("Fileset"))
        logfile.add_argument(
            "filename",  nargs="?", default="-",
            help="Local filename to be saved to. '-' for stdout")
        logopts = logfile.add_mutually_exclusive_group()
        logopts.add_argument(
            "--name", action="store_true",
            help="return the path of the logfile within the ManagedRepository")
        logopts.add_argument(
            "--size", action="store_true",
            help="return the size of the logfile in bytes")

        usage = parser.add(sub, self.usage)
        usage.set_args_unsorted()
        usage.add_login_arguments()
        usage.add_style_argument()
        usage.add_argument(
            "--wait", type=int,
            help="Number of seconds to wait for the processing to complete "
            "(Indefinite < 0; No wait=0).", default=-1)
        usage.add_argument(
            "--size-only", action="store_true",
            help="Print total bytes used in bytes")
        usage.add_argument(
            "--report", action="store_true",
            help="Print detailed breakdown of disk usage")
        usage.add_argument(
            "--sum-by", nargs="+", choices=("user", "group", "component"),
            help=("Breakdown of disk usage by a combination of "
                  "user, group and component"))
        usage.add_argument(
            "--sort-by", nargs="+",
            choices=("user", "group", "component", "size", "files"),
            help=("Sort the report table by one or more of "
                  "user, group, component, size and files"))
        usage.add_argument(
            "--reverse", action="store_true",
            help="Reverse sort order")
        unit_group = usage.add_mutually_exclusive_group()
        unit_group.add_argument(
            "--units", choices="KMGTP",
            help="Units to use for disk usage")
        unit_group.add_argument(
            "--human-readable", action="store_true",
            help="Use most appropriate units")
        usage.add_argument(
            "--groups",  action="store_true",
            help="Print size for all current user's groups")
        usage.add_argument(
            "obj", nargs="*",
            help=("Objects to be queried in the form "
                  "'<Class>:<Id>[,<Id> ...]', or '<Class>:*' "
                  "to query all the objects of the given type "))

        importtime = parser.add(sub, self.importtime)
        importtime.add_argument(
            "fileset", nargs="?",
            type=ProxyStringType("Fileset"))
        importtime_alternatives = importtime.add_mutually_exclusive_group()
        importtime_alternatives.add_argument(
            "--cache", action="store_true",
            help="to cache the results by annotating the fileset")
        importtime_alternatives.add_argument(
            "--summary", action="store_true",
            help="summarize the results cached for filesets")

        for x in (images, sets):
            x.add_argument(
                "--extended", action="store_true",
                help="provide more details for each (slow)")

    def _table(self, args):
        from omero.util.text import TableBuilder
        tb = TableBuilder("#")
        if args.style:
            tb.set_style(args.style)
        return tb

    def _extended_info(self, client, row, values):

        from omero.cmd import ManageImageBinaries

        rsp = None
        try:
            mib = ManageImageBinaries()
            mib.imageId = row[0]
            cb = client.submit(mib)
            try:
                rsp = cb.getResponse()
            finally:
                cb.close(True)
        except Exception as e:
            self.ctx.dbg("Error on MIB: %s" % e)

        if rsp is None:
            values.extend(["ERR", "ERR"])
            return  # Early exit!

        if rsp.pixelsPresent:
            values.append(filesizeformat(rsp.pixelSize))
        elif rsp.pixelSize == 0:
            values.append(filesizeformat(0))
        else:
            v = "%s (bak)" % filesizeformat(rsp.pixelSize)
            values.append(v)
        values.append(filesizeformat(rsp.pyramidSize))

    def images(self, args):
        """
        List images, filtering for archives, etc.

        This command is useful for showing pre-FS (i.e. OMERO 4.4
        and before) images which have original data archived with
        them. It *may* be possible to convert these to OMERO 5
        filesets.

        Examples:
        ```
        omero fs images --archived       # List only OMERO4 images
        omero fs images --order=newest   # Default
        omero fs images --order=largest  # Most used space
        omero fs images --limit=500      # Longer listings
        omero fs images --extended       # More details
        ```
        """

        from omero.rtypes import unwrap
        from omero.sys import ParametersI

        select = (
            "select i.id, i.name, fs.id,"
            "count(f1.id)+count(f2.id), "
            "sum(coalesce(f1.size,0) + coalesce(f2.size, 0)) ")
        archived = (not args.archived and "left outer " or "")
        query1 = (
            "from Image i join i.pixels p "
            "%sjoin p.pixelsFileMaps m %sjoin m.parent f1 "
            "left outer join i.fileset as fs "
            "left outer join fs.usedFiles as uf "
            "left outer join uf.originalFile as f2 ") % \
            (archived, archived)
        query2 = (
            "group by i.id, i.name, fs.id ")

        if args.order == "newest":
            query3 = "order by i.id desc"
        elif args.order == "oldest":
            query3 = "order by i.id asc"
        elif args.order == "largest":
            query3 = "order by "
            query3 += "sum(coalesce(f1.size,0) + coalesce(f2.size, 0)) desc"

        client = self.ctx.conn(args)
        service = client.sf.getQueryService()

        count = unwrap(service.projection(
            "select count(i) " + query1,
            None, {"omero.group": "-1"}))[0][0]
        rows = unwrap(service.projection(
            select + query1 + query2 + query3,
            ParametersI().page(args.offset, args.limit),
            {"omero.group": "-1"}))

        # Formatting
        for row in rows:
            if row[2] is None:
                row[2] = ""
            bytes = row[4]
            row[4] = filesizeformat(bytes)

        cols = ["Image", "Name", "FS", "# Files", "Size"]
        if args.extended:
            cols.extend(["Pixels", "Pyramid"])

        tb = self._table(args)
        tb.page(args.offset, args.limit, count)
        tb.cols(cols)
        for idx, row in enumerate(rows):
            values = list(row)
            if args.extended:
                self._extended_info(client, row, values)
            tb.row(idx, *tuple(values))
        self.ctx.out(str(tb.build()))

    @admin_only(AdminPrivilegeWriteManagedRepo, AdminPrivilegeChown)
    def mkdir(self, args):
        """
        Make a new directory (admin-only)

        Creates a new empty directory in the managed repository.
        A new storage volume may then be mounted at that location
        and the import template (omero.fs.repo.path) adjusted to
        target it. Once created, the directory may be deleted from
        the underlying filesystem and replaced with a symbolic link.
        Directories that violate the root-owned prefix components of
        ``omero.fs.repo.path`` are all set to be owned by the root user.
        """

        if len(args.new_dir) < 2:
            raise ValueError("directory path too short", args.new_dir)
        if args.new_dir[0] == '/':
            args.new_dir = args.new_dir[1:]
        if args.new_dir[-1] != '/':
            args.new_dir += '/'

        client = self.ctx.conn(args)

        mrepo = client.getManagedRepository()
        mrepo.makeDir(args.new_dir, args.parents)

    @windows_warning
    # Remove decorator from disabled rename to more promptly raise Exception.
    # @admin_only(AdminPrivilegeWriteOwned, AdminPrivilegeWriteManagedRepo,
    #             AdminPrivilegeDeleteOwned, AdminPrivilegeDeleteManagedRepo)
    def rename(self, args):
        """
        Moves an existing fileset to a new location (admin-only)

        After the import template (omero.fs.repo.path) has been changed,
        it may be useful to rename an existing fileset to match the new
        template. By default the original files and import log are also
        moved.
        """

        # See https://trello.com/c/J3LNquSH/ for more information.
        # When reenabling, also reenable testRenameAdminOnly.
        self.ctx.die(30, 'disabled since OMERO 5.4.7 due to Pixels.path bug')
        # Keep privilege imports used until @admin_only decorator restored.
        [AdminPrivilegeWriteOwned, AdminPrivilegeWriteManagedRepo,
         AdminPrivilegeDeleteOwned, AdminPrivilegeDeleteManagedRepo]

        fid = args.fileset.id.val
        client = self.ctx.conn(args)
        uid = self.ctx.get_event_context().userId
        isAdmin = self.ctx.get_event_context().isAdmin
        query = client.sf.getQueryService()

        try:
            fileset = query.get("Fileset", fid, {"omero.group": "-1"})
            p = fileset.details.permissions
            oid = fileset.details.owner.id.val
            gid = fileset.details.group.id.val
            if not p.canEdit():
                self.ctx.die(110, "Cannot edit Fileset:%s" % fid)
            elif oid != uid and not isAdmin:
                self.ctx.die(111, "Fileset:%s belongs to %s" % (fid, oid))
        except ServerError as se:
            self.ctx.die(
                112, "Could not load Fileset:%s- %s" % (fid, se.message))

        new_client = None
        if oid != uid:
            user = query.get("Experimenter", oid)
            group = query.get("ExperimenterGroup", gid)
            principal = Principal(
                user.omeName.val, group.name.val, "Sessions")
            service = client.sf.getSessionService()
            session = service.createSessionWithTimeouts(
                principal, 0, 30000)
            props = client.getPropertyMap()
            new_client = Client(props)
            new_client.joinSession(session.uuid.val)
            client = new_client

        tomove = []
        try:
            mrepo = client.getManagedRepository()
            root = mrepo.root()
            prefix = prep_directory(client, mrepo)
            self.ctx.err("Renaming Fileset:%s to %s" % (fid, prefix))
            tomove = rename_fileset(client, mrepo, fileset, prefix)
        finally:
            if new_client is not None:
                new_client.__del__()

        if not tomove:
            self.ctx.die(113, "No files moved!")
        elif not args.no_move:
            from omero.grid import RawAccessRequest
            for from_path, to_path in tomove:
                raw = RawAccessRequest()
                raw.repoUuid = root.hash.val
                raw.command = "mv"
                raw.args = [from_path, to_path]
                self.ctx.err("Moving %s to %s" % (from_path, to_path))
                try:
                    self.ctx.get_client().submit(raw)
                except CmdError as ce:
                    self.ctx.die(114, ce.err)
        else:
            self.ctx.err(
                "Done. You will now need to move these files manually:")
            self.ctx.err(
                "-----------------------------------------------------")
            b = "".join([root.path.val, root.name.val])
            for from_path, to_path in tomove:
                t = "/".join([b, to_path])
                f = "/".join([b, from_path])
                cmd = "mv %s %s" % (f, t)
                self.ctx.out(cmd)

    def repos(self, args):
        """
        List all repositories.

        These repositories are where OMERO stores all binary data for your
        system. Most useful is likely the ``ManagedRepository`` where OMERO 5
        imports to.

        Examples:
        ```
        omero fs repos            # Show all
        omero fs repos --managed  # Show only the managed repo or print only the directory under Unix  # noqa
        omero fs repos --managed --style=plain | cut -d, -f5
        ```
        """

        from omero.grid import ManagedRepositoryPrx as MRepo
        from functools import cmp_to_key

        def my_cmp(a, b):
            return cmp(a[0].id.val, b[0].id.val)

        client = self.ctx.conn(args)
        shared = client.sf.sharedResources()
        repos = shared.repositories()
        repos = list(zip(repos.descriptions, repos.proxies))
        repos.sort(key = cmp_to_key(my_cmp))

        tb = self._table(args)
        tb.cols(["Id", "UUID", "Type", "Path"])
        for idx, pair in enumerate(repos):
            desc, prx = pair
            path = "".join([desc.path.val, desc.name.val])

            type = "Public"
            is_mrepo = MRepo.checkedCast(prx)
            if is_mrepo:
                type = "Managed"
            if args.managed and not is_mrepo:
                continue
            if desc.hash.val == "ScriptRepo":
                type = "Script"
            tb.row(idx, *(desc.id.val, desc.hash.val, type, path))
        self.ctx.out(str(tb.build()))

    def sets(self, args):
        """
        List filesets by various criteria

        Filesets are bundles of original data imported into OMERO 5 and above
        which represent 1 *or more* images.

        Examples:

        ```
        omero fs sets --order=newest        # Default
        omero fs sets --order=oldest
        omero fs sets --order=largest
        omero fs sets --without-images      # Corrupt filesets
        omero fs sets --with-transfer=ln_s  # Symlinked filesets
        omero fs sets --check               # Proof the checksums
        ```
        """

        from omero.constants.namespaces import NSFILETRANSFER
        from omero_sys_ParametersI import ParametersI
        from omero.rtypes import unwrap

        client = self.ctx.conn(args)
        service = client.sf.getQueryService()
        admin = client.sf.getAdminService()

        if args.check and not admin.getEventContext().isAdmin:
            self.error_admin_only(fatal=True)

        annselect = (
            "(select ann.textValue from Fileset f4 "
            "join f4.annotationLinks fal join fal.child ann "
            "where f4.id = fs.id and ann.ns =:ns) ")
        select = (
            "select fs.id, fs.templatePrefix, "
            "(select size(f2.images) from Fileset f2 "
            "where f2.id = fs.id),"
            "(select size(f3.usedFiles) from Fileset f3 "
            "where f3.id = fs.id),") \
            + annselect
        query1 = (
            "from Fileset fs "
            "where 1 = 1 ")
        query2 = (
            "group by fs.id, fs.templatePrefix ")

        if args.order:
            if args.order == "newest":
                query2 += "order by fs.id desc"
            elif args.order == "oldest":
                query2 += "order by fs.id asc"
            elif args.order == "prefix":
                query2 += "order by fs.templatePrefix"

        if args.without_images:
            query = "%s and fs.images is empty %s" % (query1, query2)
        else:
            query = "%s %s" % (query1, query2)

        params = ParametersI()
        params.addString("ns", NSFILETRANSFER)
        count = service.projection("select count(fs) " + query1,
                                   params, {"omero.group": "-1"})

        params.page(args.offset, args.limit)
        objs = service.projection(select + query,
                                  params, {"omero.group": "-1"})
        objs = unwrap(objs)
        count = unwrap(count)[0][0]

        cols = ["Id", "Prefix", "Images", "Files", "Transfer"]
        if args.check:
            cols.append("Check")
        tb = self._table(args)
        tb.cols(cols)
        tb.page(args.offset, args.limit, count)

        # Map any requested transfers as well
        if args.with_transfer:
            restricted = [TRANSFERS.get(x, x) for x in args.with_transfer[0]]
        else:
            restricted = None

        for idx, obj in enumerate(objs):

            # Map the transfer name to the CLI symbols
            ns = obj[-1]
            if ns is None:
                ns = ""
            elif ns in TRANSFERS:
                ns = TRANSFERS[ns]
            obj[-1] = ns

            # Filter based on the ns symbols
            if restricted and ns not in restricted:
                continue

            # Now perform check if required
            if args.check:
                from omero.grid import RawAccessRequest
                desc, prx = self.get_managed_repo(client)
                ctx = client.getContext(group=-1)
                check_params = ParametersI()
                check_params.addId(obj[0])
                rows = service.projection((
                    "select h.value, f.hash, "
                    "f.path || '/' || f.name "
                    "from Fileset fs join fs.usedFiles uf "
                    "join uf.originalFile f join f.hasher h "
                    "where fs.id = :id"
                    ), check_params, ctx)

                if not rows:
                    obj.append("Empty")

                err = None
                for row in rows:
                    row = unwrap(row)
                    raw = RawAccessRequest()
                    raw.repoUuid = desc.hash.val
                    raw.command = "checksum"
                    raw.args = list(map(str, row))
                    try:
                        cb = client.submit(raw)
                        cb.close(True)
                    except CmdError as ce:
                        err = ce.err
                        self.ctx.dbg(err)

                if err:
                    obj.append("ERROR!")
                elif rows:
                    obj.append("OK")

            tb.row(idx, *tuple(obj))
        self.ctx.out(str(tb.build()))

    def ls(self, args):
        """
        List all the original files contained in a fileset
        """
        client = self.ctx.conn(args)
        gateway = BlitzGateway(client_obj=client)
        gateway.SERVICE_OPTS.setOmeroGroup("-1")
        fileset = gateway.getObject("Fileset", args.fileset.id.val)

        defaultdict(list)
        for ofile in fileset.listFiles():
            print(ofile.path + ofile.name)

    def logfile(self, args):
        """
        Return the logfile associated with a fileset
        """
        client = self.ctx.conn(args)
        query = client.sf.getQueryService()
        log = get_logfile(query, args.fileset.id.val)
        if log is not None:
            if args.name:
                self.ctx.out(log.path.val + log.name.val)
            elif args.size:
                self.ctx.out(log.size.val)
            else:
                target_file = str(args.filename)
                try:
                    if target_file == "-":
                        client.download(log, filehandle=sys.stdout)
                        sys.stdout.flush()
                    else:
                        client.download(log, target_file)
                except ValidationException as ve:
                    # This should effectively be handled by None being
                    # returned from the logfile query above.
                    self.ctx.die(115, "ValidationException: %s" % ve.message)
                except ResourceError as re:
                    # ID exists in DB, but not on FS
                    self.ctx.die(116, "ResourceError: %s" % re.message)
        else:
            self.ctx.die(
                117,
                "Log file not accessible for Fileset:%s" % args.fileset.id.val)

    def get_managed_repo(self, client):
        """
        For the moment this assumes there's only one.
        """
        from omero.grid import ManagedRepositoryPrx as MRepo

        shared = client.sf.sharedResources()
        repos = shared.repositories()
        repos = list(zip(repos.descriptions, repos.proxies))
        repos.sort(lambda a, b: cmp(a[0].id.val, b[0].id.val))

        for idx, pair in enumerate(repos):
            if MRepo.checkedCast(pair[1]):
                return pair

    def usage(self, args):
        """
        Shows the disk usage for various objects.

        This command shows the total disk usage of various objects including:
        ExperimenterGroup, Experimenter, Project, Dataset, Folder, Screen, Plate,
        Well, WellSample, Image, Pixels, Annotation, Job, Fileset, OriginalFile.
        The total size returned will comprise the disk usage by all related files. Thus
        an image's size would typically include the files uploaded to a fileset,
        import log (Job), thumbnails, and, possibly, associated pixels or original
        files. These details can be displayed using the --report option.

        Examples:

        ```
        omero fs usage             # total usage for current user
        omero fs usage --report    # more detailed usage for current user
        omero fs usage --groups    # total usage for current user's groups
        # total usage for five images with minimal output
        omero fs usage Image:1,2,3,4,5 --size_only
        # total usage for all images with in a human readable format
        omero fs usage Image:* --human-readable
        # total usage for all users broken down by user and group
        omero fs usage Experimenter:* --report --sum-by user group
        # total usage for two projects and one dataset Megabytes
        omero fs usage Project:1,2 Dataset:5 --units M
        # in this last case if the dataset was within project 1 or 2
        # then the size returned would be identical to:
        omero fs usage Project:1,2 --units M
        ```
        """
        from omero.cmd import DiskUsage2

        client = self.ctx.conn(args)
        req = DiskUsage2()
        if not args.obj:
            admin = client.sf.getAdminService()
            uid = admin.getEventContext().userId
            if not args.groups:
                args.obj.append("Experimenter:%d" % uid)
            else:
                exp = admin.getExperimenter(uid)
                groups = exp.linkedExperimenterGroupList()
                gids = [x.id.val for x in groups]
                args.obj.append(
                    "ExperimenterGroup:%s" % ",".join(map(str, gids)))

        req.targetObjects, req.targetClasses = self._usage_obj(args.obj)
        cb = None
        try:
            rsp, status, cb = self.response(client, req, wait=args.wait)
            self._usage_report(req, rsp, status, args)
        finally:
            if cb is not None:
                cb.close(True)  # Close handle

    def _usage_obj(self, obj):
        """
        Take the positional arguments and marshal them into
        a dictionary and a list for the command argument.
        """
        objects = {}
        classes = set()
        for o in obj:
            try:
                parts = o.split(":", 1)
                assert len(parts) == 2
                klass = parts[0]
                if '*' in parts[1]:
                    classes.add(klass)
                else:
                    ids = [int(id) for id in parts[1].split(",")]
                    objects.setdefault(klass, []).extend(ids)
            except:
                raise ValueError("Bad object: ", o)

        return (objects, list(classes))

    def _to_units(self, size, units):
        """
        Convert from bytes to KiB, MiB, GiB, TiB or PiB.
        """
        oneK = 1024.0
        powers = {'K': 1, 'M': 2, 'G': 3, 'T': 4, 'P': 5}
        if units in list(powers.keys()):
            return round(old_div(size,oneK**powers[units]), 1)
        else:
            raise ValueError("Unrecognized units: ", units)

    def _usage_report(self, req, rsp, status, args):
        """
        Output the total bytes used or the error,
        optionally provide more details.
        """
        err = self.get_error(rsp)
        if err:
            self.ctx.err(err)
        else:
            size = sum(rsp.totalBytesUsed.values())
            if args.human_readable:
                size_str = filesizeformat(size)
            else:
                size_str = "%s" % size

            if args.size_only:
                self.ctx.out(size_str)
            else:
                files = sum(rsp.totalFileCount.values())
                if args.units:
                    size_str = ("%s %siB"
                            % (self._to_units(size, args.units), args.units))
                self.ctx.out(
                    "Total disk usage: %s bytes in %d files"
                    % (size_str, files))

            if args.report and not args.size_only and size > 0:
                self._detailed_usage_report(req, rsp, status, args)

    def _detailed_usage_report(self, req, rsp, status, args):
        """
        Print a breakdown of disk usage in table form, including user,
        group and component information according to the args.
        """
        from omero.util.text import TableBuilder

        sum_by = ("user", "group", "component")
        if args.sum_by is not None:
            sum_by = args.sum_by
        showCols = list(sum_by)
        showCols.extend(["size", "files"])

        align = 'l'*len(sum_by)
        align += 'rr'
        tb = TableBuilder(*showCols)
        tb.set_align(align)
        if args.style:
            tb.set_style(args.style)

        subtotals = {}
        if "component" in sum_by:
            for userGroup in list(rsp.bytesUsedByReferer.keys()):
                for (element, size) in list(rsp.bytesUsedByReferer[
                        userGroup].items()):
                    files = rsp.fileCountByReferer[userGroup][element]
                    keyList = []
                    if "user" in sum_by:
                        keyList.append(userGroup.first)
                    if "group" in sum_by:
                        keyList.append(userGroup.second)
                    keyList.append(element)
                    key = tuple(keyList)
                    if key in list(subtotals.keys()):
                        subtotals[key][0] += size
                        subtotals[key][1] += files
                    else:
                        subtotals[key] = [size, files]
        else:
            for userGroup in list(rsp.totalBytesUsed.keys()):
                size = rsp.totalBytesUsed[userGroup]
                files = rsp.totalFileCount[userGroup]
                keyList = []
                if "user" in sum_by:
                    keyList.append(userGroup.first)
                if "group" in sum_by:
                    keyList.append(userGroup.second)
                key = tuple(keyList)
                if key in list(subtotals.keys()):
                    subtotals[key][0] += size
                    subtotals[key][1] += files
                else:
                    subtotals[key] = [size, files]

        for key in list(subtotals.keys()):
            row = list(key)
            row.extend(subtotals[key])
            tb.row(*tuple(row))

        # Since an order in the response is not guaranteed if not sort keys
        # are specified then sort by the first column at least.
        if args.sort_by:
            keys = []
            for col in args.sort_by:
                try:
                    pos = showCols.index(col)
                    keys.append(pos)
                except:
                    pass
        else:
            keys = [0]
        tb.sort(cols=keys, reverse=args.reverse)

        # Format the size column after sorting.
        if args.units:
            col = tb.get_col("size")
            col = [self._to_units(val, args.units) for val in col]
            tb.replace_col("size", col)
            tb.replace_header("size", "size (%siB)" % args.units)
        elif args.human_readable:
            col = tb.get_col("size")
            col = [filesizeformat(val) for val in col]
            tb.replace_col("size", col)
        else:
            tb.replace_header("size", "size (bytes)")

        self.ctx.out(str(tb.build()))

    def importtime(self, args):
        """Find out how long it took to import an existing fileset"""
        client = self.ctx.conn(args)
        import_time = ImportTime(self.ctx, client.sf.getQueryService())
        if args.fileset:
            if args.summary:
                self.ctx.die(28, "no summary if fileset provided")
            import_time.fileset_id = args.fileset.id
            import_time.get_cache()
            if not import_time.metrics:
                import_time.query_durations()
                import_time.query_counts()
                if args.cache:
                    import_time.write_cache(client.sf.getUpdateService())
            import_time.print_report()
        elif args.summary:
            import_time.print_summary()
        else:
            self.ctx.die(29, "provide fileset or request summary")


class ImportTime(object):

    def __init__(self, ctx, query):
        self.cli_ctx = ctx
        self.ice_ctx = {"omero.group": "-1"}
        self.query = query
        self.ns = 'openmicroscopy.org/omero/import/metrics'
        self.metrics = dict()

        import_tuples = [
            ('UPLOAD', 'upload (ms)'), ('UPLOAD_C', '# files'),
            ('SET_ID', 'setId (ms)'),
            ('METADATA', 'metadata (ms)'),
            ('PIXELDATA', 'pixeldata (ms)'), ('PIXELDATA_C', '# planes'),
            ('OVERLAY', 'overlays (ms)'),
            ('RDEF', 'rnd defs (ms)'), ('RDEF_C', '# settings'),
            ('THUMBNAIL', 'thumbnails (ms)'), ('THUMBNAIL_C', '# thumbnails')
        ]

        # Easier in Python 2.7 with OrderedDict.
        # Even easier in Python 3.7 in which dictionaries preserve ordering.
        self.import_phases = [key for (key, name) in import_tuples]
        self.import_phases_to_names = dict(import_tuples)
        self.import_names_to_phases = dict(
            [(y, x) for (x, y) in import_tuples])

    def query_durations(self):
        """Determine values for the phase durations for the import metrics"""
        from omero.sys import ParametersI

        # Get the upload job ID and its creation time, and the import log ID.

        hql = (
            "SELECT u.id, u.details.creationEvent.time, jol.child.id "
            "FROM FilesetJobLink fjl, UploadJob u, JobOriginalFileLink jol "
            "WHERE :id = fjl.parent.id AND fjl.child = u AND u = jol.parent "
            "AND jol.child.mimetype = :mimetype "
            "ORDER BY u.id"
        )

        results = self.query.projection(
            hql, ParametersI()
            .addId(self.fileset_id)
            .addString('mimetype', 'application/omero-log-file')
            .page(0, 1),
            self.ice_ctx)

        if not results:
            self.cli_ctx.die(30, 'Could not query for import log.')

        upload_job_id = results[0][0].val
        upload_start = results[0][1].val
        import_log_id = results[0][2].val

        # From the event log to find when the upload job was first updated.

        hql = (
            "SELECT event.time "
            "FROM EventLog "
            "WHERE action = :action "
            "AND entityType = :type AND entityId = :id "
            "ORDER BY id"
        )

        results = self.query.projection(
            hql, ParametersI()
            .addId(upload_job_id)
            .addString('type', 'ome.model.jobs.UploadJob')
            .addString('action', 'UPDATE')
            .page(0, 1),
            self.ice_ctx)

        if not results:
            self.cli_ctx.die(31, 'Upload job is created but not yet updated.')

        upload_end = results[0][0].val

        # Find when the import log was updated.
        # Its size is set after each step.

        hql = (
            "SELECT id, event.time "
            "FROM EventLog "
            "WHERE action = :action "
            "AND entityType = :type AND entityId = :id "
            "ORDER BY id"
        )

        results = self.query.projection(
            hql, ParametersI()
            .addId(import_log_id)
            .addString('type', 'ome.model.core.OriginalFile')
            .addString('action', 'UPDATE')
            .page(0, 3),
            self.ice_ctx)

        if not results or len(results) < 3:
            self.cli_ctx.die(32, 'Thumbnails step is not yet finished.')

        metadata_before_id = results[0][0]
        pixeldata_before_id = results[1][0]
        thumbnails_before_id = results[2][0]
        metadata_end = results[0][1].val
        pixeldata_end = results[1][1].val
        thumbnails_end = results[2][1].val

        # Find when the fileset's images were created.
        # Used as an estimate for when setId completed.
        # Ignores any inserts that follow the metadata phase.

        hql = (
            "SELECT el.event.time "
            "FROM Image i, EventLog el "
            "WHERE el.id < :last AND el.action = :action "
            "AND el.entityType = :type AND el.entityId = i.id "
            "AND i.fileset.id = :id"
        )

        results = self.query.projection(
            hql, ParametersI()
            .addId(self.fileset_id)
            .addString('type', 'ome.model.core.Image')
            .addString('action', 'INSERT')
            .add('last', metadata_before_id)
            .page(0, 1),
            self.ice_ctx)

        if not results:
            self.cli_ctx.die(33, 'Could not find images from metadata step.')

        set_id_end = results[0][0].val

        # Find when any ROIs were created during the thumbnails step.
        # These would be overlays as one finds with MIAS plates.

        hql = (
            "SELECT el.event.time "
            "FROM Roi r, EventLog el "
            "WHERE el.id > :first AND el.id < :last AND el.action = :action "
            "AND el.entityType = :type AND el.entityId = r.id "
            "AND r.image.fileset.id = :id"
        )

        results = self.query.projection(
            hql, ParametersI()
            .addId(self.fileset_id)
            .addString('type', 'ome.model.roi.Roi')
            .addString('action', 'INSERT')
            .add('first', pixeldata_before_id)
            .add('last', thumbnails_before_id)
            .page(0, 1),
            self.ice_ctx)

        overlays_start = results[0][0].val if results else None

        # Find when any rendering settings were created during the thumbnails
        # step. These are created if the thumbnails can already be generated.

        hql = (
            "SELECT el.event.time "
            "FROM RenderingDef r, EventLog el "
            "WHERE el.id > :first AND el.id < :last AND el.action = :action "
            "AND el.entityType = :type AND el.entityId = r.id "
            "AND r.pixels.image.fileset.id = :id"
        )

        results = self.query.projection(
            hql, ParametersI()
            .addId(self.fileset_id)
            .addString('type', 'ome.model.display.RenderingDef')
            .addString('action', 'INSERT')
            .add('first', pixeldata_before_id)
            .add('last', thumbnails_before_id)
            .page(0, 1),
            self.ice_ctx)

        settings_start = results[0][0].val if results else None

        # Find when any thumbnails were created during the thumbnails step.
        # These are created even if it is not yet possible to generate them.

        hql = (
            "SELECT el.event.time "
            "FROM Thumbnail t, EventLog el "
            "WHERE el.id > :first AND el.id < :last AND el.action = :action "
            "AND el.entityType = :type AND el.entityId = t.id "
            "AND t.pixels.image.fileset.id = :id"
        )

        results = self.query.projection(
            hql, ParametersI()
            .addId(self.fileset_id)
            .addString('type', 'ome.model.display.Thumbnail')
            .addString('action', 'INSERT')
            .add('first', pixeldata_before_id)
            .add('last', thumbnails_before_id)
            .page(0, 1),
            self.ice_ctx)

        thumbnails_start = results[0][0].val if results else None

        # Calculate duration of import phases.

        self.metrics['UPLOAD'] = upload_end - upload_start
        self.metrics['SET_ID'] = set_id_end - upload_end
        self.metrics['METADATA'] = metadata_end - set_id_end
        self.metrics['PIXELDATA'] = pixeldata_end - metadata_end

        if overlays_start:
            if settings_start:
                self.metrics['OVERLAY'] = settings_start - pixeldata_end
            elif thumbnails_start:
                self.metrics['OVERLAY'] = thumbnails_start - pixeldata_end
            else:
                self.metrics['OVERLAY'] = thumbnails_end - pixeldata_end

        if settings_start:
            if thumbnails_start:
                self.metrics['RDEF'] = thumbnails_start - settings_start
                self.metrics['THUMBNAIL'] = thumbnails_end - thumbnails_start
            else:
                self.metrics['RDEF'] = thumbnails_end - settings_start

    def query_counts(self):
        """Determine values for the per-item counts for the import metrics"""
        from omero.sys import ParametersI
        fileset = ParametersI().addId(self.fileset_id)

        hql = (
            "SELECT COUNT(*) FROM FilesetEntry " +
            "WHERE fileset.id = :id"
        )

        result = self.query.projection(hql, fileset,
                                       self.ice_ctx)[0][0].val
        if result > 0:
            self.metrics['UPLOAD_C'] = result

        if 'PIXELDATA' in self.metrics:
            hql = (
                "SELECT SUM(sizeC * sizeT * sizeZ) FROM Pixels " +
                "WHERE image.fileset.id = :id"
            )

            result = self.query.projection(hql, fileset,
                                           self.ice_ctx)[0][0].val
            if result > 0:
                self.metrics['PIXELDATA_C'] = result

        if 'RDEF' in self.metrics:
            hql = (
                "SELECT COUNT(*) FROM RenderingDef " +
                "WHERE pixels.image.fileset.id = :id " +
                "AND details.owner = pixels.details.owner"
                )

            result = self.query.projection(hql, fileset,
                                           self.ice_ctx)[0][0].val
            if result > 0:
                self.metrics['RDEF_C'] = result

        if 'THUMBNAIL' in self.metrics:
            hql = (
                "SELECT COUNT(*) FROM Thumbnail " +
                "WHERE pixels.image.fileset.id = :id " +
                "AND details.owner = pixels.details.owner"
                )

            result = self.query.projection(hql, fileset,
                                           self.ice_ctx)[0][0].val
            if result > 0:
                self.metrics['THUMBNAIL_C'] = result

    def get_cache(self):
        """Retrieve import metrics from a map annotation on the fileset"""
        from omero.sys import ParametersI

        hql = (
            "SELECT mv.name, mv.value "
            "FROM FilesetAnnotationLink AS l "
            "JOIN l.child.mapValue AS mv "
            "WHERE l.parent.id = :id AND l.child.ns = :ns"
        )

        results = self.query.projection(
            hql, ParametersI()
            .addId(self.fileset_id)
            .addString('ns', self.ns),
            self.ice_ctx)

        if results:
            for [name, value] in results:
                phase = self.import_names_to_phases.get(name.val)
                if phase:
                    self.metrics[phase] = int(value.val)

    def write_cache(self, update):
        """Write import metrics to a map annotation on the fileset"""
        from omero.model import FilesetI
        from omero.model import FilesetAnnotationLinkI
        from omero.model import MapAnnotationI
        from omero.model import NamedValue

        link = FilesetAnnotationLinkI()
        link.parent = FilesetI(self.fileset_id, False)
        link.child = MapAnnotationI()

        link.child.ns = rstring(self.ns)
        link.child.mapValue = []

        for phase in self.import_phases:
            if phase in self.metrics:
                link.child.mapValue.append(NamedValue(
                    self.import_phases_to_names[phase],
                    str(self.metrics[phase])))

        update.saveObject(link)

    def print_report(self):
        """Report how long it took to import an existing fileset"""
        metrics_keys = set(self.metrics)

        if set(['UPLOAD', 'UPLOAD_C']) <= metrics_keys:
            time = old_div(self.metrics['UPLOAD'], 1000.0)
            count = self.metrics['UPLOAD_C']
            plural = "s" if count > 1 else ""
            print(("   upload time of {0:6.2f}s for "
                   "{1} file{2} ({3:.3f}s/file)")
                  .format(time, count, plural, old_div(time,count)))

        time = old_div(self.metrics['SET_ID'], 1000.0)
        print("    setId time of {0:6.2f}s".format(time))

        time = old_div(self.metrics['METADATA'], 1000.0)
        print(" metadata time of {0:6.2f}s".format(time))

        if set(['PIXELDATA', 'PIXELDATA_C']) <= metrics_keys:
            time = old_div(self.metrics['PIXELDATA'], 1000.0)
            count = self.metrics['PIXELDATA_C']
            plural = "s" if count > 1 else ""
            print(("   pixels time of {0:6.2f}s for "
                   "{1} plane{2} ({3:.3f}s/plane)")
                  .format(time, count, plural, old_div(time,count)))

        if 'OVERLAY' in metrics_keys:
            time = old_div(self.metrics['OVERLAY'], 1000.0)
            print(" overlays time of {0:6.2f}s".format(time))

        if set(['RDEF', 'RDEF_C']) <= metrics_keys:
            time = old_div(self.metrics['RDEF'], 1000.0)
            count = self.metrics['RDEF_C']
            plural = "s" if count > 1 else ""
            print(("    rdefs time of {0:6.2f}s for "
                   "{1} rendering setting{2} ({3:.3f}s/rdef)")
                  .format(time, count, plural, old_div(time,count)))

        if set(['THUMBNAIL', 'THUMBNAIL_C']) <= metrics_keys:
            time = old_div(self.metrics['THUMBNAIL'], 1000.0)
            count = self.metrics['THUMBNAIL_C']
            plural = "s" if count > 1 else ""
            print(("thumbnail time of {0:6.2f}s for "
                   "{1} thumbnail{2} ({3:.3f}s/thumbnail)")
                  .format(time, count, plural, old_div(time,count)))

    def print_summary(self):
        """
        Report import metrics from map annotations on filesets
        """
        from omero.sys import ParametersI

        hql = (
            "SELECT l.parent.id, mv.name, mv.value "
            "FROM FilesetAnnotationLink AS l "
            "JOIN l.child.mapValue AS mv "
            "WHERE l.child.ns = :ns "
            "ORDER BY l.parent.id"
        )

        results = self.query.projection(
            hql, ParametersI()
            .addString('ns', self.ns),
            self.ice_ctx)

        if not results:
            print("no import times to report")
            return

        columns = ['fileset']
        for phase in self.import_phases:
            columns.append(self.import_phases_to_names[phase])
        print(','.join(['"{0}"'.format(column) for column in columns]))

        self.fileset_id = None

        for result in results:
            if self.fileset_id != result[0].val:
                self.print_summary_line()
                self.fileset_id = result[0].val
            phase = self.import_names_to_phases.get(result[1].val)
            if phase:
                self.metrics[phase] = int(result[2].val)
        self.print_summary_line()

    def print_summary_line(self):
        """
        Report import metrics from the map annotations on a fileset
        """
        if not self.metrics:
            return
        values = [str(self.fileset_id)]
        for phase in self.import_phases:
            if phase in self.metrics:
                values.append(str(self.metrics[phase]))
            else:
                values.append('')
        print(','.join(values))
        self.metrics.clear()


try:
    register("fs", FsControl, HELP)
except NameError:
    if __name__ == "__main__":
        cli = CLI()
        cli.register("fs", FsControl, HELP)
        cli.invoke(sys.argv[1:])



# -*- coding: utf-8 -*-
# **********************************************************************
#
# Copyright (c) 2003-2017 ZeroC, Inc. All rights reserved.
#
# This copy of Ice is licensed to you under the terms described in the
# ICE_LICENSE file included in this distribution.
#
# **********************************************************************
#
# Ice version 3.6.4
#
# <auto-generated>
#
# Generated from file `Exporter.ice'
#
# Warning: do not edit this file.
#
# </auto-generated>
#

from sys import version_info as _version_info_
import Ice, IcePy
import omero_ServerErrors_ice
import omero_ServicesF_ice

# Included module Ice
_M_Ice = Ice.openModule('Ice')

# Included module Glacier2
_M_Glacier2 = Ice.openModule('Glacier2')

# Included module omero
_M_omero = Ice.openModule('omero')

# Included module omero.model
_M_omero.model = Ice.openModule('omero.model')

# Included module omero.sys
_M_omero.sys = Ice.openModule('omero.sys')

# Included module omero.api
_M_omero.api = Ice.openModule('omero.api')

# Included module omero.grid
_M_omero.grid = Ice.openModule('omero.grid')

# Start of module omero
__name__ = 'omero'

# Start of module omero.api
__name__ = 'omero.api'

if 'Exporter' not in _M_omero.api.__dict__:
    _M_omero.api.Exporter = Ice.createTempClass()
    class Exporter(_M_omero.api.StatefulServiceInterface):
        """
        Stateful service for generating OME-XML or OME-TIFF from data stored
        in OMERO. Intended usage:
        {@code
        ExporterPrx e = sf.createExporter();
        // Exporter is currently in the configuration state
        // Objects can be added by id which should be present
        // in the output.
        e.addImage(1);
        // As soon as a generate method is called, the objects
        // added to the Exporter are converted to the specified
        // format. The length of the file produced is returned.
        // No more objects can be added to the Exporter, nor can
        // another generate method be called.
        long length = e.generateTiff();
        // As soon as the server-side file is generated, read()
        // can be called to get file segments. To create another
        // file, create a second Exporter. Be sure to close all
        // Exporter instances.
        long read = 0
        byte\[] buf;
        while (true) {
        buf = e.read(read, 1000000);
        // Store to file locally here
        if (buf.length
        """
        def __init__(self):
            if Ice.getType(self) == _M_omero.api.Exporter:
                raise RuntimeError('omero.api.Exporter is an abstract class')

        def ice_ids(self, current=None):
            return ('::Ice::Object', '::omero::api::Exporter', '::omero::api::ServiceInterface', '::omero::api::StatefulServiceInterface')

        def ice_id(self, current=None):
            return '::omero::api::Exporter'

        def ice_staticId():
            return '::omero::api::Exporter'
        ice_staticId = staticmethod(ice_staticId)

        def addImage_async(self, _cb, id, current=None):
            """
            Adds a single image with basic metadata to the Exporter for inclusion
            on the next call to getBytes().
            Arguments:
            _cb -- The asynchronous callback object.
            id -- 
            current -- The Current object for the invocation.
            """
            pass

        def generateXml_async(self, _cb, current=None):
            """
            Generates an OME-XML file. The return value is the length
            of the file produced.
            Arguments:
            _cb -- The asynchronous callback object.
            current -- The Current object for the invocation.
            """
            pass

        def generateTiff_async(self, _cb, current=None):
            """
            Generates an OME-TIFF file. The return value is the length
            of the file produced. This method ends configuration.
            Arguments:
            _cb -- The asynchronous callback object.
            current -- The Current object for the invocation.
            """
            pass

        def read_async(self, _cb, position, length, current=None):
            """
            Returns length bytes from the output file. The
            file can be safely read until reset() is called.
            Arguments:
            _cb -- The asynchronous callback object.
            position -- 
            length -- 
            current -- The Current object for the invocation.
            """
            pass

        def __str__(self):
            return IcePy.stringify(self, _M_omero.api._t_Exporter)

        __repr__ = __str__

    _M_omero.api.ExporterPrx = Ice.createTempClass()
    class ExporterPrx(_M_omero.api.StatefulServiceInterfacePrx):

        """
        Adds a single image with basic metadata to the Exporter for inclusion
        on the next call to getBytes().
        Arguments:
        id -- 
        _ctx -- The request context for the invocation.
        """
        def addImage(self, id, _ctx=None):
            return _M_omero.api.Exporter._op_addImage.invoke(self, ((id, ), _ctx))

        """
        Adds a single image with basic metadata to the Exporter for inclusion
        on the next call to getBytes().
        Arguments:
        id -- 
        _response -- The asynchronous response callback.
        _ex -- The asynchronous exception callback.
        _sent -- The asynchronous sent callback.
        _ctx -- The request context for the invocation.
        Returns: An asynchronous result object for the invocation.
        """
        def begin_addImage(self, id, _response=None, _ex=None, _sent=None, _ctx=None):
            return _M_omero.api.Exporter._op_addImage.begin(self, ((id, ), _response, _ex, _sent, _ctx))

        """
        Adds a single image with basic metadata to the Exporter for inclusion
        on the next call to getBytes().
        Arguments:
        id -- 
        """
        def end_addImage(self, _r):
            return _M_omero.api.Exporter._op_addImage.end(self, _r)

        """
        Generates an OME-XML file. The return value is the length
        of the file produced.
        Arguments:
        _ctx -- The request context for the invocation.
        """
        def generateXml(self, _ctx=None):
            return _M_omero.api.Exporter._op_generateXml.invoke(self, ((), _ctx))

        """
        Generates an OME-XML file. The return value is the length
        of the file produced.
        Arguments:
        _response -- The asynchronous response callback.
        _ex -- The asynchronous exception callback.
        _sent -- The asynchronous sent callback.
        _ctx -- The request context for the invocation.
        Returns: An asynchronous result object for the invocation.
        """
        def begin_generateXml(self, _response=None, _ex=None, _sent=None, _ctx=None):
            return _M_omero.api.Exporter._op_generateXml.begin(self, ((), _response, _ex, _sent, _ctx))

        """
        Generates an OME-XML file. The return value is the length
        of the file produced.
        Arguments:
        """
        def end_generateXml(self, _r):
            return _M_omero.api.Exporter._op_generateXml.end(self, _r)

        """
        Generates an OME-TIFF file. The return value is the length
        of the file produced. This method ends configuration.
        Arguments:
        _ctx -- The request context for the invocation.
        """
        def generateTiff(self, _ctx=None):
            return _M_omero.api.Exporter._op_generateTiff.invoke(self, ((), _ctx))

        """
        Generates an OME-TIFF file. The return value is the length
        of the file produced. This method ends configuration.
        Arguments:
        _response -- The asynchronous response callback.
        _ex -- The asynchronous exception callback.
        _sent -- The asynchronous sent callback.
        _ctx -- The request context for the invocation.
        Returns: An asynchronous result object for the invocation.
        """
        def begin_generateTiff(self, _response=None, _ex=None, _sent=None, _ctx=None):
            return _M_omero.api.Exporter._op_generateTiff.begin(self, ((), _response, _ex, _sent, _ctx))

        """
        Generates an OME-TIFF file. The return value is the length
        of the file produced. This method ends configuration.
        Arguments:
        """
        def end_generateTiff(self, _r):
            return _M_omero.api.Exporter._op_generateTiff.end(self, _r)

        """
        Returns length bytes from the output file. The
        file can be safely read until reset() is called.
        Arguments:
        position -- 
        length -- 
        _ctx -- The request context for the invocation.
        """
        def read(self, position, length, _ctx=None):
            return _M_omero.api.Exporter._op_read.invoke(self, ((position, length), _ctx))

        """
        Returns length bytes from the output file. The
        file can be safely read until reset() is called.
        Arguments:
        position -- 
        length -- 
        _response -- The asynchronous response callback.
        _ex -- The asynchronous exception callback.
        _sent -- The asynchronous sent callback.
        _ctx -- The request context for the invocation.
        Returns: An asynchronous result object for the invocation.
        """
        def begin_read(self, position, length, _response=None, _ex=None, _sent=None, _ctx=None):
            return _M_omero.api.Exporter._op_read.begin(self, ((position, length), _response, _ex, _sent, _ctx))

        """
        Returns length bytes from the output file. The
        file can be safely read until reset() is called.
        Arguments:
        position -- 
        length -- 
        """
        def end_read(self, _r):
            return _M_omero.api.Exporter._op_read.end(self, _r)

        def checkedCast(proxy, facetOrCtx=None, _ctx=None):
            return _M_omero.api.ExporterPrx.ice_checkedCast(proxy, '::omero::api::Exporter', facetOrCtx, _ctx)
        checkedCast = staticmethod(checkedCast)

        def uncheckedCast(proxy, facet=None):
            return _M_omero.api.ExporterPrx.ice_uncheckedCast(proxy, facet)
        uncheckedCast = staticmethod(uncheckedCast)

        def ice_staticId():
            return '::omero::api::Exporter'
        ice_staticId = staticmethod(ice_staticId)

    _M_omero.api._t_ExporterPrx = IcePy.defineProxy('::omero::api::Exporter', ExporterPrx)

    _M_omero.api._t_Exporter = IcePy.defineClass('::omero::api::Exporter', Exporter, -1, (), True, False, None, (_M_omero.api._t_StatefulServiceInterface,), ())
    Exporter._ice_type = _M_omero.api._t_Exporter

    Exporter._op_addImage = IcePy.Operation('addImage', Ice.OperationMode.Normal, Ice.OperationMode.Normal, True, None, (), (((), IcePy._t_long, False, 0),), (), None, (_M_omero._t_ServerError,))
    Exporter._op_generateXml = IcePy.Operation('generateXml', Ice.OperationMode.Normal, Ice.OperationMode.Normal, True, None, (), (), (), ((), IcePy._t_long, False, 0), (_M_omero._t_ServerError,))
    Exporter._op_generateTiff = IcePy.Operation('generateTiff', Ice.OperationMode.Normal, Ice.OperationMode.Normal, True, None, (), (), (), ((), IcePy._t_long, False, 0), (_M_omero._t_ServerError,))
    Exporter._op_read = IcePy.Operation('read', Ice.OperationMode.Idempotent, Ice.OperationMode.Idempotent, True, None, (), (((), IcePy._t_long, False, 0), ((), IcePy._t_int, False, 0)), (), ((), _M_Ice._t_ByteSeq, False, 0), (_M_omero._t_ServerError,))

    _M_omero.api.Exporter = Exporter
    del Exporter

    _M_omero.api.ExporterPrx = ExporterPrx
    del ExporterPrx

# End of module omero.api

__name__ = 'omero'

# End of module omero
